"""
Core model handling routines.
"""
from __future__ import print_function

__all__ = [
    "list_models", "load_model", "load_model_info",
    "build_model", "precompile_dlls", "reparameterize",
    ]

import os
from os.path import basename, join as joinpath
from glob import glob
import re
import copy

import numpy as np # type: ignore

# NOTE: delay loading of kernelcl, kernelcuda, kerneldll and kernelpy
# cl and cuda in particular take awhile since they try to establish a
# connection with the card to verify that the environment works.
from . import generate
from . import modelinfo
from . import product
from . import mixture
from . import custom

# pylint: disable=unused-import
try:
    from typing import List, Union, Optional, Any, Tuple
    from .kernel import KernelModel
    from .modelinfo import ModelInfo
except ImportError:
    pass
# pylint: enable=unused-import

CUSTOM_MODEL_PATH = os.environ.get('SAS_MODELPATH', "")
if CUSTOM_MODEL_PATH == "":
    CUSTOM_MODEL_PATH = joinpath(os.path.expanduser("~"), ".sasmodels", "custom_models")
    #if not os.path.isdir(CUSTOM_MODEL_PATH):
    #    os.makedirs(CUSTOM_MODEL_PATH)

# TODO: refactor composite model support
# The current load_model_info/build_model does not reuse existing model
# definitions when loading a composite model, instead reloading and
# rebuilding the kernel for each component model in the expression.  This
# is fine in a scripting environment where the model is built when the script
# starts and is thrown away when the script ends, but may not be the best
# solution in a long-lived application.  This affects the following functions:
#
#    load_model
#    load_model_info
#    build_model

KINDS = ("all", "py", "c", "double", "single", "opencl", "1d", "2d",
         "nonmagnetic", "magnetic")
def list_models(kind=None):
    # type: (str) -> List[str]
    """
    Return the list of available models on the model path.

    *kind* can be one of the following:

        * all: all models
        * py: python models only
        * c: c models only
        * single: c models which support single precision
        * double: c models which require double precision
        * opencl: c models which run in opencl
        * dll: c models which do not run in opencl
        * 1d: models without orientation
        * 2d: models with orientation
        * magnetic: models supporting magnetic sld
        * nommagnetic: models without magnetic parameter

    For multiple conditions, combine with plus.  For example, *c+single+2d*
    would return all oriented models implemented in C which can be computed
    accurately with single precision arithmetic.
    """
    if kind and any(k not in KINDS for k in kind.split('+')):
        raise ValueError("kind not in " + ", ".join(KINDS))
    files = sorted(glob(joinpath(generate.MODEL_PATH, "[a-zA-Z]*.py")))
    available_models = [basename(f)[:-3] for f in files]
    if kind and '+' in kind:
        all_kinds = kind.split('+')
        condition = lambda name: all(_matches(name, k) for k in all_kinds)
    else:
        condition = lambda name: _matches(name, kind)
    selected = [name for name in available_models if condition(name)]

    return selected

def _matches(name, kind):
    if kind is None or kind == "all":
        return True
    info = load_model_info(name)
    pars = info.parameters.kernel_parameters
    # TODO: may be adding Fq to the list at some point
    is_pure_py = callable(info.Iq)
    if kind == "py":
        return is_pure_py
    elif kind == "c":
        return not is_pure_py
    elif kind == "double":
        return not info.single and not is_pure_py
    elif kind == "single":
        return info.single and not is_pure_py
    elif kind == "opencl":
        return info.opencl
    elif kind == "dll":
        return not info.opencl and not is_pure_py
    elif kind == "2d":
        return any(p.type == 'orientation' for p in pars)
    elif kind == "1d":
        return all(p.type != 'orientation' for p in pars)
    elif kind == "magnetic":
        return any(p.type == 'sld' for p in pars)
    elif kind == "nonmagnetic":
        return not any(p.type == 'sld' for p in pars)
    return False

def load_model(model_name, dtype=None, platform='ocl'):
    # type: (str, str, str) -> KernelModel
    """
    Load model info and build model.

    *model_name* is the name of the model, or perhaps a model expression
    such as sphere*hardsphere or sphere+cylinder.

    *dtype* and *platform* are given by :func:`build_model`.
    """
    return build_model(load_model_info(model_name),
                       dtype=dtype, platform=platform)

def load_model_info(model_string):
    # type: (str) -> modelinfo.ModelInfo
    """
    Load a model definition given the model name.

    *model_string* is the name of the model, or perhaps a model expression
    such as sphere*cylinder or sphere+cylinder. Use '@' for a structure
    factor product, e.g. sphere@hardsphere. Custom models can be specified by
    prefixing the model name with 'custom.', e.g. 'custom.MyModel+sphere'.

    This returns a handle to the module defining the model.  This can be
    used with functions in generate to build the docs or extract model info.
    """
    if "+" in model_string:
        parts = [load_model_info(part)
                 for part in model_string.split("+")]
        return mixture.make_mixture_info(parts, operation='+')
    elif "*" in model_string:
        parts = [load_model_info(part)
                 for part in model_string.split("*")]
        return mixture.make_mixture_info(parts, operation='*')
    elif "@" in model_string:
        p_info, q_info = [load_model_info(part)
                          for part in model_string.split("@")]
        return product.make_product_info(p_info, q_info)
    # We are now dealing with a pure model
    elif "custom." in model_string:
        pattern = "custom.([A-Za-z0-9_-]+)"
        result = re.match(pattern, model_string)
        if result is None:
            raise ValueError("Model name in invalid format: " + model_string)
        model_name = result.group(1)
        # Use ModelName to find the path to the custom model file
        model_path = joinpath(CUSTOM_MODEL_PATH, model_name + ".py")
        if not os.path.isfile(model_path):
            raise ValueError("The model file {} doesn't exist".format(model_path))
        kernel_module = custom.load_custom_kernel_module(model_path)
        return modelinfo.make_model_info(kernel_module)
    kernel_module = generate.load_kernel_module(model_string)
    return modelinfo.make_model_info(kernel_module)


_REPARAMETERIZE_DOCS = """\
Definition
----------

Constrain :ref:`%(base)s` according to the following::

    %(translation)s
"""
_LHS_RE = re.compile(r"^ *(?<![.0-9])([A-Za-z_][A-Za-z0-9_]+) *=",
                     flags=re.MULTILINE)
def reparameterize(
        base, parameters, translation, filename=None,
        title=None, insert_after=None, docs=None, name=None,
        source=None,
    ):
    """
    Reparameterize an existing model.

    *base* is the original modelinfo. This cannot be a reparameterized model;
    only one level of reparameterization is supported.

    *parameters* are the new parameter definitions that will be
    included in the model info.

    *translation* is a string each line containing *var = expr*.  The variable
    *var* can be a new intermediate value, or it can be a parameter from
    the base model that will be replace by the expression.  The expression
    *expr* can be any C99 expression, including C-style if-expressions
    *condition ? value1 : value2*.  Expressions can use any new or existing
    parameter that is not being replaced including intermediate values that
    are previously defined.  Parameters can only be assigned once, never
    updated.  C99 math functions are available, as well as any functions
    defined in the base model or included in *source* (see below).

    *filename* is the filename for the replacement model.  This is usually
    *__file__*, giving the path to the model file, but it could also be a
    nominal filename for translations defined on-the-fly.

    *title* is the model title, which defaults to *base.title* plus
    " (reparameterized)".

    *insert_after* controls parameter placement.  By default, the new
    parameters replace the old parameters in their original position.
    Instead, you can provide a dictionary *{'par': 'newpar1,newpar2'}*
    indicating that new parameters named *newpar1* and *newpar2* should
    be included in the table after the existing parameter *par*, or at
    the beginning if *par* is the empty string.

    *docs* constains the doc string for the translated model, which by default
    references the base model and gives the *translation* text.

    *name* is the model name (default = :code:`"constrained_" + base.name`).

    *source* is a list any additional C source files that should be included to
    define functions and constants used in the translation expressions.  This
    will be included after all sources for the base model.  Sources will only
    be included once, even if they are listed in both places, so feel free to
    list all dependencies for the helper function, such as "lib/polevl.c".
    """
    if not isinstance(base, modelinfo.ModelInfo):
        base = load_model_info(base)
    if name is None:
        name = filename if filename is not None else "constrained_" + base.name
    name = os.path.basename(name).split('.')[0]
    if title is None:
        title = base.title + " (reparameterized)"
    if docs is None:
        lines = "\n    ".join(s.lstrip() for s in translation.split('\n'))
        docs = _REPARAMETERIZE_DOCS%{'base': base.id, 'translation': lines}
    #source = merge_deps(base.source, source)
    source = (base.source + [f for f in source if f not in base.source]
              if source else base.source)


    # TODO: don't repeat code from generate._build_translation
    base_pars = [par.id for par in base.parameters.kernel_parameters]
    old_pars = [match.group(1) for match in _LHS_RE.finditer(translation)
                if match.group(1) in base_pars]
    new_pars = [modelinfo.parse_parameter(*p) for p in parameters]
    table = modelinfo.derive_table(base.parameters, remove=old_pars,
                                   insert=new_pars, insert_after=insert_after)

    caller = copy.copy(base)
    caller.translation = translation
    caller.name = caller.id = name
    caller.docs = docs
    caller.filename = filename
    caller.parameters = table
    caller.source = source
    return caller

# Note: not used at the moment.
def merge_deps(old, new):
    """
    Merge two dependency lists.  The lists are partially ordered, with
    all dependents coming after the items they depend on, but otherwise
    order doesn't matter.  The merged list preserves the partial ordering.
    So if old and new both include the item "c", then all items that come
    before "c" in old and new will come before "c" in the result, and all
    items that come after "c" in old and new will come after "c" in the
    result.
    """
    if new is None:
        return old

    result = []
    for item in new:
        try:
            index = old.index(item)
            #print(item,"found in",old,"at",index,"giving",old[:index])
            result.extend(old[:index])
            old = old[index+1:]
        except ValueError:
            #print(item, "not found in", old)
            pass
        result.append(item)
        #print("after", item, "old", old, "result", result)
    result.extend(old)
    return result


def build_model(model_info, dtype=None, platform="ocl"):
    # type: (ModelInfo, str, str) -> KernelModel
    """
    Prepare the model for the default execution platform.

    This will return an OpenCL model, a DLL model or a python model depending
    on the model and the computing platform.

    *model_info* is the model definition structure returned from
    :func:`load_model_info`.

    *dtype* indicates whether the model should use single or double precision
    for the calculation.  Choices are 'single', 'double', 'quad', 'half',
    or 'fast'.  If *dtype* ends with '!', then force the use of the DLL rather
    than OpenCL for the calculation.

    *platform* should be "dll" to force the dll to be used for C models,
    otherwise it uses the default "ocl".
    """
    composition = model_info.composition
    if composition is not None:
        composition_type, parts = composition
        models = [build_model(p, dtype=dtype, platform=platform) for p in parts]
        if composition_type == 'mixture':
            return mixture.MixtureModel(model_info, models)
        elif composition_type == 'product':
            P, S = models
            return product.ProductModel(model_info, P, S)
        else:
            raise ValueError('unknown mixture type %s'%composition_type)

    # If it is a python model, return it immediately
    if callable(model_info.Iq):
        from . import kernelpy
        return kernelpy.PyModel(model_info)

    numpy_dtype, fast, platform = parse_dtype(model_info, dtype, platform)
    source = generate.make_source(model_info)
    if platform == "dll":
        from . import kerneldll
        #print("building dll", numpy_dtype)
        return kerneldll.load_dll(source['dll'], model_info, numpy_dtype)
    elif platform == "cuda":
        from . import kernelcuda
        return kernelcuda.GpuModel(source, model_info, numpy_dtype, fast=fast)
    else:
        from . import kernelcl
        #print("building ocl", numpy_dtype)
        return kernelcl.GpuModel(source, model_info, numpy_dtype, fast=fast)

def precompile_dlls(path, dtype="double"):
    # type: (str, str) -> List[str]
    """
    Precompile the dlls for all builtin models, returning a list of dll paths.

    *path* is the directory in which to save the dlls.  It will be created if
    it does not already exist.

    This can be used when build the windows distribution of sasmodels
    which may be missing the OpenCL driver and the dll compiler.
    """
    from . import kerneldll

    numpy_dtype = np.dtype(dtype)
    if not os.path.exists(path):
        os.makedirs(path)
    compiled_dlls = []
    for model_name in list_models():
        model_info = load_model_info(model_name)
        if not callable(model_info.Iq):
            source = generate.make_source(model_info)['dll']
            old_path = kerneldll.SAS_DLL_PATH
            try:
                kerneldll.SAS_DLL_PATH = path
                dll = kerneldll.make_dll(source, model_info,
                                         dtype=numpy_dtype, system=True)
            finally:
                kerneldll.SAS_DLL_PATH = old_path
            compiled_dlls.append(dll)
    return compiled_dlls

def parse_dtype(model_info, dtype=None, platform=None):
    # type: (ModelInfo, str, str) -> Tuple[np.dtype, bool, str]
    """
    Interpret dtype string, returning np.dtype, fast flag and platform.

    Possible types include 'half', 'single', 'double' and 'quad'.  If the
    type is 'fast', then this is equivalent to dtype 'single' but using
    fast native functions rather than those with the precision level
    guaranteed by the OpenCL standard.  'default' will choose the appropriate
    default for the model and platform.

    Platform preference can be specfied ("ocl", "cuda", "dll"), with the
    default being OpenCL or CUDA if available, otherwise DLL.  If the dtype
    name ends with '!' then platform is forced to be DLL rather than GPU.
    The default platform is set by the environment variable SAS_OPENCL,
    SAS_OPENCL=driver:device for OpenCL, SAS_OPENCL=cuda:device for CUDA
    or SAS_OPENCL=none for DLL.

    This routine ignores the preferences within the model definition.  This
    is by design.  It allows us to test models in single precision even when
    we have flagged them as requiring double precision so we can easily check
    the performance on different platforms without having to change the model
    definition.
    """
    # Assign default platform, overriding ocl with dll if OpenCL is unavailable
    # If opencl=False OpenCL is switched off
    if platform is None:
        platform = "ocl"

    # Check if type indicates dll regardless of which platform is given
    if dtype is not None and dtype.endswith('!'):
        platform = "dll"
        dtype = dtype[:-1]

    # Make sure model allows opencl/gpu
    if not model_info.opencl:
        platform = "dll"

    # Make sure opencl is available, or fallback to cuda then to dll
    if platform == "ocl":
        from . import kernelcl
        if not kernelcl.use_opencl():
            from . import kernelcuda
            platform = "cuda" if kernelcuda.use_cuda() else "dll"

    # Convert special type names "half", "fast", and "quad"
    fast = (dtype == "fast")
    if fast:
        dtype = "single"
    elif dtype == "quad":
        dtype = "longdouble"
    elif dtype == "half":
        dtype = "float16"

    # Convert dtype string to numpy dtype.  Use single precision for GPU
    # if model allows it, otherwise use double precision.
    if dtype is None or dtype == "default":
        numpy_dtype = (generate.F32 if model_info.single and platform in ("ocl", "cuda")
                       else generate.F64)
    else:
        numpy_dtype = np.dtype(dtype)

    # Make sure that the type is supported by GPU, otherwise use dll
    if platform == "ocl":
        from . import kernelcl
        env = kernelcl.environment()
    elif platform == "cuda":
        from . import kernelcuda
        env = kernelcuda.environment()
    else:
        env = None
    if env is not None and not env.has_type(numpy_dtype):
        platform = "dll"
        if dtype is None:
            numpy_dtype = generate.F64

    return numpy_dtype, fast, platform

def test_composite_order():
    """
    Check that mixture models produce the same result independent of ordder.
    """
    def test_models(fst, snd):
        """Confirm that two models produce the same parameters"""
        fst = load_model(fst)
        snd = load_model(snd)
        # Un-disambiguate parameter names so that we can check if the same
        # parameters are in a pair of composite models. Since each parameter in
        # the mixture model is tagged as e.g., A_sld, we ought to use a
        # regex subsitution s/^[A-Z]+_/_/, but removing all uppercase letters
        # is good enough.
        # TODO: check that the models produce the same results
        # Note that compare.py will give a misleading answer. For
        # "cylinder+sphere" the A_radius parameter will use the default
        # cylinder radius, but for "sphere+cylinder" it will use the default
        # sphere radius so a simple comparison of the two will appear to be
        # different unless you explicitly set radius, solvent, and solvent_sld
        # for the A and B models.
        fst = [[x for x in p.name if x == x.lower()]
               for p in fst.info.parameters.kernel_parameters]
        snd = [[x for x in p.name if x == x.lower()]
               for p in snd.info.parameters.kernel_parameters]
        assert sorted(fst) == sorted(snd), "{} != {}".format(fst, snd)

    test_models(
        "cylinder+sphere",
        "sphere+cylinder")
    test_models(
        "cylinder*sphere",
        "sphere*cylinder")
    test_models(
        "cylinder@hardsphere*sphere",
        "sphere*cylinder@hardsphere")
    test_models(
        "barbell+sphere*cylinder@hardsphere",
        "sphere*cylinder@hardsphere+barbell")
    test_models(
        "barbell+cylinder@hardsphere*sphere",
        "cylinder@hardsphere*sphere+barbell")
    test_models(
        "barbell+sphere*cylinder@hardsphere",
        "barbell+cylinder@hardsphere*sphere")
    test_models(
        "sphere*cylinder@hardsphere+barbell",
        "cylinder@hardsphere*sphere+barbell")
    test_models(
        "barbell+sphere*cylinder@hardsphere",
        "cylinder@hardsphere*sphere+barbell")
    test_models(
        "barbell+cylinder@hardsphere*sphere",
        "sphere*cylinder@hardsphere+barbell")

def test_composite():
    # type: () -> None
    """Check that model load works"""
    from .product import RADIUS_ID, VOLFRAC_ID, STRUCTURE_MODE_ID, RADIUS_MODE_ID
    #Test the the model produces the parameters that we would expect
    model = load_model("cylinder@hardsphere*sphere")
    actual = [p.name for p in model.info.parameters.kernel_parameters]
    a_parts = ("sld", "sld_solvent", "radius", "length", "theta", "phi",
        RADIUS_ID, VOLFRAC_ID, STRUCTURE_MODE_ID, RADIUS_MODE_ID)
    b_parts = ("sld", "sld_solvent", "radius")
    target = [*(f"A_{p}" for p in a_parts), *(f"B_{p}" for p in b_parts)]
    assert target == actual, "%s != %s"%(target, actual)


def list_models_main():
    # type: () -> int
    """
    Run list_models as a main program.  See :func:`list_models` for the
    kinds of models that can be requested on the command line.
    """
    import sys
    kind = sys.argv[1] if len(sys.argv) > 1 else "all"
    try:
        models = list_models(kind)
        print("\n".join(models))
    except Exception:
        print(list_models.__doc__)
        return 1
    return 0

if __name__ == "__main__":
    list_models_main()



"""
SAS model constructor.

Small angle scattering models are defined by a set of kernel functions:

    *Iq(q, p1, p2, ...)* returns the scattering at q for a form with
    particular dimensions averaged over all orientations.

    *Iqac(qab, qc, p1, p2, ...)* returns the scattering at qab, qc
    for a rotationally symmetric form with particular dimensions.
    qab, qc are determined from shape orientation and scattering angles.
    This call is used if the shape has orientation parameters theta and phi.

    *Iqabc(qa, qb, qc, p1, p2, ...)* returns the scattering at qa, qb, qc
    for a form with particular dimensions.  qa, qb, qc are determined from
    shape orientation and scattering angles. This call is used if the shape
    has orientation parameters theta, phi and psi.

    *Iqxy(qx, qy, p1, p2, ...)* returns the scattering at qx, qy.  Use this
    to create an arbitrary 2D theory function, needed for q-dependent
    background functions and for models with non-uniform magnetism.

    *form_volume(p1, p2, ...)* returns the volume of the form with particular
    dimension, or 1.0 if no volume normalization is required.

    *shell_volume(p1, p2, ...)* returns the volume of the shell for forms
    which are hollow.

    *radius_effective(mode, p1, p2, ...)* returns the effective radius of
    the form with particular dimensions.  Mode determines the type of
    effective radius returned, with mode=1 for equivalent volume.

These functions are defined in a kernel module .py script and an associated
set of .c files.  The model constructor will use them to create models with
polydispersity across volume and orientation parameters, and provide
scale and background parameters for each model.

C code should be stylized C-99 functions written for OpenCL. All functions
need prototype declarations even if the are defined before they are used.
Although OpenCL supports *#include* preprocessor directives, the list of
includes should be given as part of the metadata in the kernel module
definition. The included files should be listed using a path relative to the
kernel module, or if using "lib/file.c" if it is one of the standard includes
provided with the sasmodels source. The includes need to be listed in order
so that functions are defined before they are used.

Floating point values should be declared as *double*.  For single precision
calculations, *double* will be replaced by *float*.  The single precision
conversion will also tag floating point constants with "f" to make them
single precision constants.  When using integral values in floating point
expressions, they should be expressed as floating point values by including
a decimal point.  This includes 0., 1. and 2.

OpenCL has a *sincos* function which can improve performance when both
the *sin* and *cos* values are needed for a particular argument.  Since
this function does not exist in C99, all use of *sincos* should be
replaced by the macro *SINCOS(value, sn, cn)* where *sn* and *cn* are
previously declared *double* variables.  When compiled for systems without
OpenCL, *SINCOS* will be replaced by *sin* and *cos* calls.   If *value* is
an expression, it will appear twice in this case; whether or not it will be
evaluated twice depends on the quality of the compiler.

The kernel module must set variables defining the kernel meta data:

    *id* is an implicit variable formed from the filename.  It will be
    a valid python identifier, and will be used as the reference into
    the html documentation, with '_' replaced by '-'.

    *name* is the model name as displayed to the user.  If it is missing,
    it will be constructed from the id.

    *title* is a short description of the model, suitable for a tool tip,
    or a one line model summary in a table of models.

    *description* is an extended description of the model to be displayed
    while the model parameters are being edited.

    *parameters* is the list of parameters.  Parameters in the kernel
    functions must appear in the same order as they appear in the
    parameters list.  Two additional parameters, *scale* and *background*
    are added to the beginning of the parameter list.  They will show up
    in the documentation as model parameters, but they are never sent to
    the kernel functions.  Note that *effect_radius* and *volfraction*
    must occur first in structure factor calculations.

    *category* is the default category for the model.  The category is
    two level structure, with the form "group:section", indicating where
    in the manual the model will be located.  Models are alphabetical
    within their section.

    *source* is the list of C-99 source files that must be joined to
    create the OpenCL kernel functions.  The files defining the functions
    need to be listed before the files which use the functions.

    *form_volume*, *Iq*, *Iqac*, *Iqabc* are strings containing
    the C source code for the body of the volume, Iq, and Iqac functions
    respectively.  These can also be defined in the last source file.

    *Iq*, *Iqac*, *Iqabc* also be instead be python functions defining the
    kernel.  If they are marked as *Iq.vectorized = True* then the
    kernel is passed the entire *q* vector at once, otherwise it is
    passed values one *q* at a time.  The performance improvement of
    this step is significant.

    *valid* expression that evaluates to True if the input parameters
    are valid (e.g., "bell_radius >= radius" for the barbell or capped
    cylinder models).  The expression can call C functions, including
    those defined in your model file.

A :class:`.modelinfo.ModelInfo` structure is constructed from the kernel meta
data and returned to the caller.

Valid inputs should be identified by the *valid* expression.
Particularly with polydispersity, there are some sets of shape parameters
which lead to nonsensical forms, such as a capped cylinder where the
cap radius is smaller than the cylinder radius.  The polydispersity
calculation will ignore these points, effectively chopping the
parameter weight distributions at the boundary of the infeasible region.
The resulting scattering will be set to background, even for models with
no polydispersity.  If the *valid* expression misses some parameter
combinations and they reach the kernel, the kernel should probably return
NaN rather than zero.  Even if the volume also evaluates to zero for
these parameters, the distribution weights are still accumulated and the
average volume calculation will be slightly off.

The doc string at the start of the kernel module will be used to
construct the model documentation web pages.  Embedded figures should
appear in the subdirectory "img" beside the model definition, and tagged
with the kernel module name to avoid collision with other models.  Some
file systems are case-sensitive, so only use lower case characters for
file names and extensions.

Code follows the C99 standard with the following extensions and conditions::

    M_PI_180 = pi/180
    M_4PI_3 = 4pi/3
    square(x) = x*x
    cube(x) = x*x*x
    sas_sinx_x(x) = sin(x)/x, with sin(0)/0 -> 1
    all double precision constants must include the decimal point
    all double declarations may be converted to half, float, or long double
    FLOAT_SIZE is the number of bytes in the converted variables

:func:`load_kernel_module` loads the model definition file and
:func:`.modelinfo.make_model_info` parses it. :func:`make_source`
converts C-based model definitions to C source code, including the
polydispersity integral.  :func:`model_sources` returns the list of
source files the model depends on, and :func:`ocl_timestamp` returns
the latest time stamp amongst the source files (so you can check if
the model needs to be rebuilt).

The function :func:`make_doc` extracts the doc string and adds the
parameter table to the top.  *make_figure* in *sasmodels/doc/genmodel*
creates the default figure for the model.  [These two sets of code
should mignrate into docs.py so docs can be updated in one place].
"""
from __future__ import print_function

# TODO: determine which functions are useful outside of generate
#__all__ = ["model_info", "make_doc", "make_source", "convert_type"]

import sys
from os import environ
from os.path import abspath, dirname, join as joinpath, exists, getmtime, sep
import re
import string
from zlib import crc32
from inspect import currentframe, getframeinfo
import logging

import numpy as np  # type: ignore

from .modelinfo import Parameter
from .custom import load_custom_kernel_module

# pylint: disable=unused-import
try:
    from typing import Tuple, Sequence, Iterator, Dict, List
    from types import ModuleType
    from .modelinfo import ModelInfo
except ImportError:
    pass
# pylint: enable=unused-import

logger = logging.getLogger(__name__)

# jitter projection to use in the kernel code.  See explore/jitter.py
# for details.  To change it from a program, set generate.PROJECTION.
PROJECTION = 1

def get_data_path(external_dir, target_file):
    """
    Search for the target file relative in the installed application.

    Search first in the location of the generate module in case we are
    running directly from the distribution.  Search next to the python
    executable for windows installs.  Search in the ../Resources directory
    next to the executable for Mac OS/X installs.
    """
    path = abspath(dirname(__file__))
    if exists(joinpath(path, target_file)):
        return path

    # check next to exe/zip file
    exepath = dirname(sys.executable)
    path = joinpath(exepath, external_dir)
    if exists(joinpath(path, target_file)):
        return path

    # check in py2app Contents/Resources
    path = joinpath(exepath, '..', 'Resources', external_dir)
    if exists(joinpath(path, target_file)):
        return abspath(path)

    raise RuntimeError('Could not find '+joinpath(external_dir, target_file))

EXTERNAL_DIR = 'sasmodels-data'
DATA_PATH = get_data_path(EXTERNAL_DIR, 'kernel_iq.c')
MODEL_PATH = joinpath(DATA_PATH, 'models')

F16 = np.dtype('float16')
F32 = np.dtype('float32')
F64 = np.dtype('float64')
try:  # CRUFT: older numpy does not support float128
    F128 = np.dtype('float128')
except TypeError:
    F128 = None

# Conversion from units defined in the parameter table for each model
# to units displayed in the sphinx documentation.
# This section associates the unit with the macro to use to produce the LaTex
# code.  The macro itself needs to be defined in sasmodels/doc/rst_prolog.
#
# NOTE: there is an RST_PROLOG at the end of this file which is NOT
# used for the bundled documentation. Still as long as we are defining the
# macros in two places any new addition should define the macro in both places.
RST_UNITS = {
    "Ang": "|Ang|",
    "1/Ang": "|Ang^-1|",
    "1/Ang^2": "|Ang^-2|",
    "Ang^3": "|Ang^3|",
    "Ang^2": "|Ang^2|",
    "1e15/cm^3": "|1e15cm^3|",
    "Ang^3/mol": "|Ang^3|/mol",
    "1e-6/Ang^2": "|1e-6Ang^-2|",
    "degrees": "degree",
    "1/cm": "|cm^-1|",
    "Ang/cm": "|Ang*cm^-1|",
    "g/cm^3": "|g/cm^3|",
    "mg/m^2": "|mg/m^2|",
    "": "None",
    }

# Headers for the parameters tables in th sphinx documentation
PARTABLE_HEADERS = [
    "Parameter",
    "Description",
    "Units",
    "Default value",
    ]

# Minimum width for a default value (this is shorter than the column header
# width, so will be ignored).
PARTABLE_VALUE_WIDTH = 10

# Documentation header for the module, giving the model name, its short
# description and its parameter table.  The remainder of the doc comes
# from the module docstring.
DOC_HEADER = """.. _%(id)s:

%(name)s
=======================================================

%(title)s

%(parameters)s

%(returns)s

%(docs)s
"""


def set_integration_size(info, n):
    # type: (ModelInfo, int) -> None
    """
    Update the model definition, replacing the gaussian integration with
    a gaussian integration of a different size.

    Note: this really ought to be a method in modelinfo, but that leads to
    import loops.
    """
    if info.source and any(lib.startswith('lib/gauss') for lib in info.source):
        from .gengauss import gengauss
        path = joinpath(MODEL_PATH, "lib", "gauss%d.c"%n)
        if not exists(path):
            gengauss(n, path)
        info.source = ["lib/gauss%d.c"%n if lib.startswith('lib/gauss')
                       else lib for lib in info.source]

def format_units(units):
    # type: (str) -> str
    """
    Convert units into ReStructured Text format.
    """
    return "string" if isinstance(units, list) else RST_UNITS.get(units, units)


def make_partable(pars):
    # type: (List[Parameter]) -> str
    """
    Generate the parameter table to include in the sphinx documentation.
    """
    column_widths = [
        max(len(p.name) for p in pars),
        max(len(p.description) for p in pars),
        max(len(format_units(p.units)) for p in pars),
        PARTABLE_VALUE_WIDTH,
        ]
    column_widths = [max(w, len(h))
                     for w, h in zip(column_widths, PARTABLE_HEADERS)]

    underbar = " ".join("="*w for w in column_widths)
    lines = [
        underbar,
        " ".join("%-*s" % (w, h)
                 for w, h in zip(column_widths, PARTABLE_HEADERS)),
        underbar,
        ]
    for p in pars:
        lines.append(" ".join([
            "%-*s" % (column_widths[0], p.name),
            "%-*s" % (column_widths[1], p.description),
            "%-*s" % (column_widths[2], format_units(p.units)),
            "%*g" % (column_widths[3], p.default),
            ]))
    lines.append(underbar)
    return "\n".join(lines)


def _search(search_path, filename):
    # type: (List[str], str) -> str
    """
    Find *filename* in *search_path*.

    Raises ValueError if file does not exist.
    """
    for path in search_path:
        target = joinpath(path, filename)
        if exists(target):
            return target
    raise ValueError("%r not found in %s" % (filename, search_path))


def model_sources(model_info):
    # type: (ModelInfo) -> List[str]
    """
    Return a list of the sources file paths for the module.
    """
    search_path = [dirname(model_info.basefile), MODEL_PATH]
    # If it is a reparameterized model, include both to directory for the
    # original model file and the directory for the reparameterized model file
    # on the search path for dependent source files.
    if model_info.filename and model_info.filename is not model_info.basefile:
        search_path.insert(0, dirname(model_info.filename))
    return [_search(search_path, f) for f in model_info.source]


def dll_timestamp(model_info):
    # type: (ModelInfo) -> int
    """
    Return a timestamp for the model corresponding to the most recently
    changed file or dependency.
    """
    # TODO: what about on-the-fly derived models---tag name with crc32
    # TODO: no longer used, and appears to be identical to ocl_timestamp
    # TODO: fails DRY; templates appear two places.
    model_templates = [joinpath(DATA_PATH, filename)
                       for filename in ('kernel_header.c', 'kernel_iq.c')]
    source_files = (model_sources(model_info)
                    + model_templates
                    + [model_info.basefile])
    # Include filename if different from basefile.
    if model_info.filename and model_info.filename is not model_info.basefile:
        source_files.append(model_info.filename)
    # Note: file may not exist when it is a standard model from library.zip
    # or when the model is generated dynamically.
    times = [getmtime(f) for f in source_files if f and exists(f)]
    newest = max(times) if times else 0
    return newest

def ocl_timestamp(model_info):
    # type: (ModelInfo) -> int
    """
    Return a timestamp for the model corresponding to the most recently
    changed file or dependency.

    Note that this does not look at the time stamps for the OpenCL header
    information since that need not trigger a recompile of the DLL.
    """
    # TODO: fails DRY; templates appear two places.
    model_templates = [joinpath(DATA_PATH, filename)
                       for filename in ('kernel_header.c', 'kernel_iq.c')]
    source_files = (model_sources(model_info)
                    + model_templates
                    + [model_info.basefile])
    # Include filename if different from basefile.
    if model_info.filename and model_info.filename is not model_info.basefile:
        source_files.append(model_info.filename)
    # Note: file may not exist when it is a standard model from library.zip
    # or when the model is generated dynamically.
    times = [getmtime(f) for f in source_files if exists(f)]
    newest = max(times) if times else 0
    return newest

def tag_source(source):
    # type: (str) -> str
    """
    Return a unique tag for the source code.
    """
    # Note: need 0xffffffff&val to force an unsigned 32-bit number
    try:
        source = source.encode('utf8')
    except AttributeError: # bytes has no encode attribute in python 3
        pass
    return "%08X"%(0xffffffff&crc32(source))

def convert_type(source, dtype):
    # type: (str, np.dtype) -> str
    """
    Convert code from double precision to the desired type.

    Floating point constants are tagged with 'f' for single precision or 'L'
    for long double precision.
    """
    source = _fix_tgmath_int(source)
    if dtype == F16:
        fbytes = 2
        source = _convert_type(source, "half", "f")
    elif dtype == F32:
        fbytes = 4
        source = _convert_type(source, "float", "f")
    elif dtype == F64:
        fbytes = 8
        # no need to convert if it is already double
    elif dtype == F128:
        fbytes = 16
        source = _convert_type(source, "long double", "L")
    else:
        raise ValueError("Unexpected dtype in source conversion: %s" % dtype)
    return ("#define FLOAT_SIZE %d\n" % fbytes)+source


def _convert_type(source, type_name, constant_flag):
    # type: (str, str, str) -> str
    """
    Replace 'double' with *type_name* in *source*, tagging floating point
    constants with *constant_flag*.
    """
    # Convert double keyword to float/long double/half.
    # Accept an 'n' # parameter for vector # values, where n is 2, 4, 8 or 16.
    # Assume complex numbers are represented as cdouble.
    # TODO: use lookahead so that "typedef complex double cdouble;" converts.
    source = re.sub(r'(^|[^a-zA-Z0-9_]c?)double(([248]|16)?($|[^a-zA-Z0-9_]))',
                    r'\1%s\2'%type_name, source)
    source = _tag_float(source, constant_flag)
    return source

TGMATH_INT_RE = re.compile(r"""
(?: # Non-capturing match; not lookbehind since pattern length is variable
  \b              # word boundary
   # various math functions
  (a?(sin|cos|tan)h? | atan2
   | erfc? | tgamma
   | exp(2|10|m1)? | log(2|10|1p)? | pow[nr]? | sqrt | rsqrt | rootn
   | fabs | fmax | fmin
   )
  \s*[(]\s*       # open parenthesis
)
[+-]?(0|[1-9]\d*) # integer
(?=               # lookahead match: don't want to move from end of int
  \s*[,)]         # comma or close parenthesis for end of argument
)                 # end lookahead
""", re.VERBOSE)
def _fix_tgmath_int(source):
    # type: (str) -> str
    """
    Replace f(integer) with f(integer.) for sin, cos, pow, etc.

    OS X OpenCL complains that it can't resolve the type generic calls to
    the standard math functions when they are called with integer constants,
    but this does not happen with the Windows Intel driver for example.
    To avoid confusion on the matrix marketplace, automatically promote
    integers to floats if we recognize them in the source.

    The specific functions we look for are:

        trigonometric: sin, asin, sinh, asinh, etc., and atan2
        exponential:   exp, exp2, exp10, expm1, log, log2, log10, logp1
        power:         pow, pown, powr, sqrt, rsqrt, rootn
        special:       erf, erfc, tgamma
        float:         fabs, fmin, fmax

    Note that we don't convert the second argument of dual argument
    functions: atan2, fmax, fmin, pow, powr.  This could potentially
    be a problem for pow(x, 2), but that case seems to work without change.
    """
    out = TGMATH_INT_RE.sub(r'\g<0>.', source)
    return out


# Floating point regular expression
#
# Define parts:
#
#    E = [eE][+-]?\d+    : Exponent
#    P = [.]             : Decimal separator
#    N = [1-9]\d*        : Natural number, no leading zeros
#    Z = 0               : Zero
#    F = \d+             : Fractional number, maybe leading zeros
#    F? = \d*            : Optional fractional number
#
# We want to reject bare natural numbers and bare decimal points, so we
# need to tediously outline the cases where we have either a fraction or
# an exponent:
#
#   ( ZP | ZPF | ZE | ZPE | ZPFE | NP | NPF | NE | NPE | NPFE | PF | PFE )
#
#
# We can then join cases by making parts optional.  The following are
# some ways to do this:
#
#   ( (Z|N)(P|PF|E|PE|PFE) | PFE? )                   # Split on lead
#     => ( (Z|N)(PF?|(PF?)?E) | PFE? )
#   ( ((Z|N)PF?|PF)E? | (Z|N)E)                       # Split on point
#   ( (ZP|ZPF|NP|NPF|PF) | (Z|ZP|ZPF|N|NP|NPF|PF)E )  # Split on E
#     => ( ((Z|N)PF?|PF) | ((Z|N)(PF?)? | PF) E )
FLOAT_RE = re.compile(r"""
    (?<!\w)  # use negative lookbehind since '.' confuses \b test
    # use split on lead to match float ( (Z|N)(PF?|(PF?)?E) | PFE? )
    ( ( 0 | [1-9]\d* )                     # ( ( Z | N )
      ([.]\d* | ([.]\d*)? [eE][+-]?\d+ )   #   (PF? | (PF?)? E )
    | [.]\d+ ([eE][+-]?\d+)?               # | PF (E)?
    )                                      # )
    (?!\w)  # use negative lookahead since '.' confuses \b test
    """, re.VERBOSE)
def _tag_float(source, constant_flag):
    # Convert floating point constants to single by adding 'f' to the end,
    # or long double with an 'L' suffix.  OS/X complains if you don't do this.
    out = FLOAT_RE.sub(r'\g<0>%s'%constant_flag, source)
    #print("in",repr(source),"out",repr(out), constant_flag)
    return out

def test_tag_float():
    """Check that floating point constants are identified and tagged with 'f'"""

    cases = """
ZP  : 0.
ZPF : 0.0,0.01,0.1
Z  E: 0e+001
ZP E: 0.E0
ZPFE: 0.13e-031
NP  : 1., 12.
NPF : 1.0001, 1.1, 1.0
N  E: 1e0, 37E-080
NP E: 1.e0, 37.E-080
NPFE: 845.017e+22
 PF : .1, .0, .0100
 PFE: .6e+9, .82E-004
# isolated cases
0.
1e0
0.13e-013
# untouched
struct3.e3, 03.05.67, 37
# expressions
3.75+-1.6e-7-27+13.2
a3.e2 - 0.
4*atan(1)
4.*atan(1.)
"""

    output = """
ZP  : 0.f
ZPF : 0.0f,0.01f,0.1f
Z  E: 0e+001f
ZP E: 0.E0f
ZPFE: 0.13e-031f
NP  : 1.f, 12.f
NPF : 1.0001f, 1.1f, 1.0f
N  E: 1e0f, 37E-080f
NP E: 1.e0f, 37.E-080f
NPFE: 845.017e+22f
 PF : .1f, .0f, .0100f
 PFE: .6e+9f, .82E-004f
# isolated cases
0.f
1e0f
0.13e-013f
# untouched
struct3.e3, 03.05.67, 37
# expressions
3.75f+-1.6e-7f-27+13.2f
a3.e2 - 0.f
4*atan(1)
4.f*atan(1.f)
"""

    for case_in, case_out in zip(cases.split('\n'), output.split('\n')):
        out = _tag_float(case_in, 'f')
        assert case_out == out, "%r => %r"%(case_in, out)


def kernel_name(model_info, variant):
    # type: (ModelInfo, str) -> str
    """
    Name of the exported kernel symbol.

    *variant* is "Iq", "Iqxy" or "Imagnetic".
    """
    return model_info.name + "_" + variant


def indent(s, depth):
    # type: (str, int) -> str
    """
    Indent a string of text with *depth* additional spaces on each line.
    """
    spaces = " "*depth
    interline_separator = "\n" + spaces
    return spaces + interline_separator.join(s.split("\n"))


_template_cache = {}  # type: Dict[str, Tuple[int, str, str]]
def load_template(filename):
    # type: (str) -> str
    """
    Load template file from sasmodels resource directory.
    """
    path = joinpath(DATA_PATH, filename)
    mtime = getmtime(path)
    if filename not in _template_cache or mtime > _template_cache[filename][0]:
        with open(path) as fid:
            _template_cache[filename] = (mtime, fid.read(), path)
    return _template_cache[filename][1], path


_FN_TEMPLATE = """\
double %(name)s(%(pars)s);
double %(name)s(%(pars)s) {
#line %(line)d "%(filename)s"
    %(body)s
}

"""
def _gen_fn(model_info, name, pars):
    # type: (ModelInfo, str, List[Parameter]) -> str
    """
    Generate a function given pars and body.

    Returns the following string::

         double fn(double a, double b, ...);
         double fn(double a, double b, ...) {
             ....
         }
    """
    par_decl = ', '.join(p.as_function_argument() for p in pars) if pars else 'void'
    body = getattr(model_info, name)
    filename = model_info.basefile
    # Note: if symbol is defined strangely in the module then default it to 1
    lineno = model_info.lineno.get(name, 1)
    return _FN_TEMPLATE % {
        'name': name, 'pars': par_decl, 'body': body,
        'filename': _clean_source_filename(filename), 'line': lineno,
    }


def _call_pars(pars, subs):
    # type: (str, List[Parameter]) -> List[str]
    """
    Return a list of *prefix+parameter* from parameter items.

    *pars* is the list of parameters from the base model.

    *subs* contains the translation equations with references to parameters
    from the new parameter table.  If there is no translation, then *subs*
    is just a list of references into the base table.
    """
    return [subs[p.id] for p in pars]

def _split_translation(translation):
    r"""
    Process the *translation* string, which is a sequence of assignments.

    Blanks and comments (c-style and python-style) are stripped.

    Conditional expressions should use C syntax (! || && ? :) not python.
    """
    # TODO: support translation for vector parameters
    # TODO: security concerns when code stored in model file
    assignments = []
    for line in translation.split('\n'):
        # strip comments and white space from line
        code = line.split('#', 1)[0].split('//', 1)[0].strip()
        if not code:
            continue
        # split assignment on first equals
        parts = code.split('=', 1)
        if len(parts) == 1:
            raise ValueError("translation expected 'var=expr' but got %r"%line)
        # store parts without assessing whether they are correct
        var, expr = parts
        assignments.append((var.strip(), expr.strip()))
    return assignments

# The phrase "(?<![.0-9])" in the following re indicates negative look-behind,
# designed to reject the like 314159e-5 and 1.0f
_IDENT_RE = re.compile(r"(?<![.0-9])([A-Za-z_][A-Za-z0-9_]*)")

def _build_translation(model_info, table_id="_v", var_prefix="_var_"):
    """
    Interpret parameter translation block, if any.

    *model_info* contains the parameter table and any translation definition
    for converting between model parameters and the base calculation model.

    *table_id* is the internal label used for the parameter call table.  It
    must be of the form "_table" matching whatever table variable name that
    appears in the macros such as CALL_VOLUME() and CALL_IQ().

    *var_prefix* is a tag to attach to intermediate variables to avoid
    collision with variables used inside kernel_iq.

    Returns:

    * *subs* = {name: expr, ...} parameter substitution table for calling
      into the kernel function
    * *translation* = "#define TRANSLATION_VARS(_v) _var_name = expr ...."
    * *validity* = "#define VALID(_v) ..."

    The returned *subs* is used to generate the substitions for CALL_VOLUME
    etc., parameters, via :func:`_call_pars`.

    The returned *translation* and *validity* macros need to be included inside
    the generated model files.  They are the same for all variants (1D, 2D,
    magnetic) so can be defined once along side the parameter table.  Even
    though they are expanded independently in each variant.
    """
    # Usual case of no translation.
    call_table = model_info.parameters.kernel_parameters
    if not model_info.translation:
        # Simple substitution {name: table_id.name}, an empty
        # TRANSLATION_VARS() macro and VALID() macro with simple subs.
        subs = {p.id: table_id+'.'+p.id for p in call_table}
        translation_vars = _build_translation_vars(table_id, [])
        validity = _build_validity_check(model_info.valid, table_id, subs)
        return subs, translation_vars, validity

    # Translation case.
    # Grab the list of assignment statements.
    assigns = _split_translation(model_info.translation)

    # Identify parameters and intermediates.  Parameters are from the calling
    # table, not the base table.  Only these have values defined by the caller
    # that can be used in expressions.  Need the list of base parameters to
    # identify substitutions.  And LHS variables that are not base parameters
    # are therefore intermediate parameters.  Note: leaving the main parameter
    # table as the caller parameter table because the rest of the code uses it
    # to build up and interpret the calling sequence.  The kernel functions
    # themselves of course use the base table.
    base_table = model_info.base.kernel_parameters
    call_pars = set(p.id for p in call_table)
    base_pars = set(p.id for p in base_table)
    variables = set(name for name, expr in assigns
                    if name not in call_pars and name not in base_pars)

    # Regular expression substition to tag the variables on each RHS.
    # Don't want to translate all symbols since they include C constants and
    # math functions.
    def id_sub(match_obj):
        name = match_obj.group(0)
        if name in variables:
            return var_prefix + name
        if name in call_pars:
            return table_id + '.' + name
        return name
    assigns = [(name, _IDENT_RE.sub(id_sub, eq)) for name, eq in assigns]

    # Create a substituion table assuming all parameters are directly assigned,
    # then update it with
    subs = {p.id: table_id+'.'+p.id for p in base_table}
    subs.update((name, eq) for name, eq in assigns if name in base_pars)

    # Build TRANSLATION_VARS macros assignments to variables.
    var_values = [(var_prefix+name, eq) for name, eq in assigns
                  if name in variables]
    translation_vars = _build_translation_vars(table_id, var_values)

    # Build VALID macro, with expressions using caller parameters.
    validity = _build_validity_check(model_info.valid, table_id, subs)

    return subs, translation_vars, validity

def _build_translation_vars(table_id, variables):
    r"""
    Build TRANSLATION_VARS macro for C which builds intermediate values.

    E. g.,

    ::

        #define TRANSLATION_VARS(_v) \
         const double _temporary_Re = cbrt(_v.volume/_v.eccentricity/M_4PI_3)
    """
    if variables:
        # Leave comma off def since last def doesn't have comma
        defs = ["const double %s = %s" % (name, eq)
                for name, eq in variables]
        return ("#define TRANSLATION_VARS(%s) \\\n  "%table_id
                + "; \\\n  ".join(defs))
    else:
        return '#define TRANSLATION_VARS(%s) do {} while(0)'%table_id

def _build_validity_check(eq, table_id, subs):
    """
    Substitute parameter expressions into validity test, returning the
    VALID(_table) macro.
    """
    if not eq:  # None or empty string
        return "#define VALID(%s) 1" % (table_id, )
    else:
        # TODO: maybe replace not/and/or with ! && ||
        # TODO: (hard) replace (a if b else c) with (b ? a : c)
        # TODO: reject a**n, or (hard) replace with pow(a, n)
        def id_sub(match_obj):
            name = match_obj.group(0)
            return "(%s)"%subs[name] if name in subs else name
        eq = _IDENT_RE.sub(id_sub, eq)
        return "#define VALID(%s) (%s)" % (table_id, eq)

_IQXY_PATTERN = re.compile(r"(^|\s)double\s+I(?P<mode>q(ac|abc|xy))\s*[(]",
                           flags=re.MULTILINE)
def find_xy_mode(source):
    # type: (List[str]) -> bool
    """
    Return the xy mode as qa, qac, qabc or qxy.

    Note this is not a C parser, and so can be easily confused by
    non-standard syntax.  Also, it will incorrectly identify the following
    as having 2D models::

        /*
        double Iqac(qab, qc, ...) { ... fill this in later ... }
        */

    If you want to comment out the function, use // on the front of the
    line::

        /*
        // double Iqac(qab, qc, ...) { ... fill this in later ... }
        */

    """
    for code in source:
        m = _IQXY_PATTERN.search(code)
        if m is not None:
            return m.group('mode')
    return 'qa'

# Note: not presently used.  Need to know whether Fq is available before
# trying to compile the source.  Leave the code here in case we decide that
# define have_Fq for each form factor is too tedious and error prone.
_FQ_PATTERN = re.compile(r"(^|\s)void\s+Fq[(]", flags=re.MULTILINE)
def contains_Fq(source):
    # type: (List[str]) -> bool
    """
    Return True if C source defines "void Fq(".
    """
    for code in source:
        if _FQ_PATTERN.search(code) is not None:
            return True
    return False

_SHELL_VOLUME_PATTERN = re.compile(r"(^|\s)double\s+shell_volume[(]",
                                   flags=re.MULTILINE)
def contains_shell_volume(source):
    # type: (List[str]) -> bool
    """
    Return True if C source defines "double shell_volume(".
    """
    for code in source:
        if _SHELL_VOLUME_PATTERN.search(code) is not None:
            return True
    return False

def _clean_source_filename(path):
    # type: (str) -> str
    """ Make the source filename into a canonical, relative form if possible

    Remove the common start of the file path (if there is one), yielding
    the path relative to this file, such as:

     - ./kernel_iq.c
     - ./models/sphere.c
     - ./models/lib/sas_J0.c

    This is a format that the compiler/debugger understand for indicating
    included files with relative paths. Omitting the common parent to the
    paths means that the irrelevant detail of the temporary directory where
    the source was unpacked for compilation is not included in pre-compiled
    models.
    """
    base = dirname(__file__).replace('\\', '/')
    path = path.replace('\\', '/')
    return path.replace(base, '.', 1) if path.startswith(base) else path

def _add_source(source, code, path, lineno=1):
    """
    Add a file to the list of source code chunks, tagged with path and line.
    """
    path = _clean_source_filename(path)
    source.append('#line %d "%s"' % (lineno, path))
    source.append(code)

def read_text(f):
    with open(f) as fid:
        return fid.read()

def make_source(model_info):
    # type: (ModelInfo) -> Dict[str, str]
    """
    Generate the OpenCL/ctypes kernel from the module info.

    Uses source files found in the given search path.  Returns None if this
    is a pure python model, with no C source components.
    """
    if callable(model_info.Iq):
        raise ValueError("can't compile python model")
        #return None

    # TODO: need something other than volume to indicate dispersion parameters
    # No volume normalization despite having a volume parameter.
    # Thickness is labelled a volume in order to trigger polydispersity.
    # May want a separate dispersion flag, or perhaps a separate category for
    # disperse, but not volume.  Volume parameters also use relative values
    # for the distribution rather than the absolute values used by angular
    # dispersion.  Need to be careful that necessary parameters are available
    # for computing volume even if we allow non-disperse volume parameters.

    # reference to table of calling parameters and table of called
    # parameters, which may be different if the model is reparameterized
    # by the caller.
    call_table = model_info.parameters
    base_table = model_info.base

    # Load templates and user code
    kernel_header = load_template('kernel_header.c')
    kernel_code = load_template('kernel_iq.c')
    user_code = [(f, read_text(f)) for f in model_sources(model_info)]

    # Build initial sources
    source = []
    _add_source(source, *kernel_header)
    for path, code in user_code:
        _add_source(source, code, path)
    if model_info.c_code:
        _add_source(source, model_info.c_code, model_info.basefile,
                    lineno=model_info.lineno.get('c_code', 1))

    # Make parameters for q, qx, qy so that we can use them in declarations
    q, qx, qy, qab, qa, qb, qc \
        = [Parameter(name=v) for v in 'q qx qy qab qa qb qc'.split()]

    # Generate form_volume, etc., function definitions from strings
    # given in the model.py file as the C function body only, such as
    # one might for the volume of a sphere:
    #     form_volume="return M_4PI_3*cube(radius)"
    if isinstance(model_info.form_volume, str):
        pars = base_table.form_volume_parameters
        source.append(_gen_fn(model_info, 'form_volume', pars))
    if isinstance(model_info.shell_volume, str):
        pars = base_table.form_volume_parameters
        source.append(_gen_fn(model_info, 'shell_volume', pars))
    if isinstance(model_info.Iq, str):
        pars = [q] + base_table.iq_parameters
        source.append(_gen_fn(model_info, 'Iq', pars))
    if isinstance(model_info.Iqxy, str):
        pars = [qx, qy] + base_table.iq_parameters + base_table.orientation_parameters
        source.append(_gen_fn(model_info, 'Iqxy', pars))
    if isinstance(model_info.Iqac, str):
        pars = [qab, qc] + base_table.iq_parameters
        source.append(_gen_fn(model_info, 'Iqac', pars))
    if isinstance(model_info.Iqabc, str):
        pars = [qa, qb, qc] + base_table.iq_parameters
        source.append(_gen_fn(model_info, 'Iqabc', pars))

    # Check for shell_volume function in source code
    is_hollow = contains_shell_volume(source)

    # What kind of 2D model do we need?  Is it consistent with the parameters?
    xy_mode = find_xy_mode(source)
    if xy_mode == 'qabc' and not base_table.is_asymmetric:
        raise ValueError("asymmetric oriented models need to define Iqabc")
    elif xy_mode == 'qac' and base_table.is_asymmetric:
        raise ValueError("symmetric oriented models need to define Iqac")
    elif not base_table.orientation_parameters and xy_mode in ('qac', 'qabc'):
        raise ValueError("Unexpected function I%s for unoriented shape"%xy_mode)
    elif base_table.orientation_parameters and xy_mode not in ('qac', 'qabc'):
        if xy_mode == 'qxy':
            logger.warning("oriented shapes should define Iqac or Iqabc")
        else:
            raise ValueError("Expected Iqac or Iqabc for oriented shape")

    # Process parameter substitutions
    subs, translation_vars, valid = _build_translation(model_info, '_v')

    # Validity check macro.  Putting it here so that line number
    # jiggering is short-lived.
    _add_source(source, valid, model_info.basefile,
                lineno=model_info.lineno.get('valid', 1))
    source.append(valid)

    # Define the parameter table
    lineno = getframeinfo(currentframe()).lineno + 2
    source.append('#line %d "sasmodels/generate.py"'%lineno)
    #source.append('introduce breakage in generate to test lineno reporting')
    source.append("#define PARAMETER_TABLE \\")
    source.append("\\\n".join(p.as_definition()
                              for p in call_table.kernel_parameters))

    # Define the function calls
    call_radius_effective = "#define CALL_RADIUS_EFFECTIVE(_mode, _v) 0.0"
    if base_table.form_volume_parameters:
        refs = _call_pars(base_table.form_volume_parameters, subs)
        if is_hollow:
            call_volume = (
                "#define CALL_VOLUME(_form, _shell, _v) "
                "do { _form = form_volume(%s); _shell = shell_volume(%s); } "
                "while (0)") % ((",".join(refs),)*2)
        else:
            call_volume = (
                "#define CALL_VOLUME(_form, _shell, _v) "
                "do { _form = _shell = form_volume(%s); } "
                "while (0)") % (",".join(refs))
        if model_info.radius_effective_modes:
            call_radius_effective = (
                "#define CALL_RADIUS_EFFECTIVE(_mode, _v) "
                "radius_effective(_mode, %s)") % (",".join(refs))
    else:
        # Model doesn't have volume.  We could make the kernel run a little
        # faster by not using/transferring the volume normalizations, but
        # the ifdef's reduce readability more than is worthwhile.
        call_volume = (
            "#define CALL_VOLUME(_form, _shell, _v) "
            "do { _form = _shell = 1.0; } while (0)")
    source.append(translation_vars)
    source.append(call_volume)
    source.append(call_radius_effective)
    model_refs = _call_pars(base_table.iq_parameters, subs)

    if model_info.have_Fq:
        pars = ",".join(["_q", "&_F1", "&_F2",] + model_refs)
        call_iq = "#define CALL_FQ(_q, _F1, _F2, _v) Fq(%s)" % pars
        clear_iq = "#undef CALL_FQ"
    else:
        pars = ",".join(["_q"] + model_refs)
        call_iq = "#define CALL_IQ(_q, _v) Iq(%s)" % pars
        clear_iq = "#undef CALL_IQ"
    if xy_mode == 'qabc':
        pars = ",".join(["_qa", "_qb", "_qc"] + model_refs)
        call_iqxy = "#define CALL_IQ_ABC(_qa,_qb,_qc,_v) Iqabc(%s)" % pars
        clear_iqxy = "#undef CALL_IQ_ABC"
    elif xy_mode == 'qac':
        pars = ",".join(["_qa", "_qc"] + model_refs)
        call_iqxy = "#define CALL_IQ_AC(_qa,_qc,_v) Iqac(%s)" % pars
        clear_iqxy = "#undef CALL_IQ_AC"
    elif xy_mode == 'qa' and not model_info.have_Fq:
        pars = ",".join(["_qa"] + model_refs)
        call_iqxy = "#define CALL_IQ_A(_qa,_v) Iq(%s)" % pars
        clear_iqxy = "#undef CALL_IQ_A"
    elif xy_mode == 'qa' and model_info.have_Fq:
        pars = ",".join(["_qa", "&_F1", "&_F2",] + model_refs)
        # Note: uses rare C construction (expr1, expr2) which computes
        # expr1 then expr2 and evaluates to expr2.  This allows us to
        # leave it looking like a function even though it is returning
        # its values by reference.
        call_iqxy = "#define CALL_FQ_A(_qa,_F1,_F2,_v) (Fq(%s),_F2)" % pars
        clear_iqxy = "#undef CALL_FQ_A"
    elif xy_mode == 'qxy':
        qxy_refs = _call_pars(base_table.orientation_parameters, subs)
        pars = ",".join(["_qx", "_qy"] + model_refs + qxy_refs)
        call_iqxy = "#define CALL_IQ_XY(_qx,_qy,_v) Iqxy(%s)" % pars
        clear_iqxy = "#undef CALL_IQ_XY"
        if base_table.orientation_parameters:
            call_iqxy += "\n#define HAVE_THETA"
            clear_iqxy += "\n#undef HAVE_THETA"
        if base_table.is_asymmetric:
            call_iqxy += "\n#define HAVE_PSI"
            clear_iqxy += "\n#undef HAVE_PSI"


    # TODO: test magnetism with translated sld parameters
    # By inspection it should work but intermediate variables aren't supported.
    magpars = [k-2 for k, p in enumerate(call_table.call_parameters)
               if p.type == 'sld']
    # Fill in definitions for numbers of parameters
    source.append("#define MAX_PD %s"%call_table.max_pd)
    source.append("#define NUM_PARS %d"%call_table.npars)
    source.append("#define NUM_VALUES %d" % call_table.nvalues)
    source.append("#define NUM_MAGNETIC %d" % call_table.nmagnetic)
    source.append("#define MAGNETIC_PARS %s"%",".join(str(k) for k in magpars))
    source.append("#define PROJECTION %d"%PROJECTION)
    wrappers = _kernels(kernel_code, call_iq, clear_iq,
                        call_iqxy, clear_iqxy, model_info.name)
    code = '\n'.join(source + wrappers[0] + wrappers[1] + wrappers[2])

    # Note: Identical code for dll and opencl.  This may not always be the case
    # so leave them as separate entries in the returned value.
    result = {'dll': code, 'opencl': code}
    return result


def _kernels(kernel, call_iq, clear_iq, call_iqxy, clear_iqxy, name):
    # type: (Dict[str, str], str, str, str, str, str) -> List[str]
    code = kernel[0]
    path = _clean_source_filename(kernel[1])
    iq = [
        # define the Iq kernel
        "#define KERNEL_NAME %s_Iq" % name,
        call_iq,
        '#line 1 "%s Iq"' % path,
        code,
        clear_iq,
        "#undef KERNEL_NAME",
        ]

    iqxy = [
        # define the Iqxy kernel from the same source with different #defines
        "#define KERNEL_NAME %s_Iqxy" % name,
        call_iqxy,
        '#line 1 "%s Iqxy"' % path,
        code,
        clear_iqxy,
        "#undef KERNEL_NAME",
    ]

    imagnetic = [
        # define the Imagnetic kernel
        "#define KERNEL_NAME %s_Imagnetic" % name,
        "#define MAGNETIC 1",
        call_iqxy,
        '#line 1 "%s Imagnetic"' % path,
        code,
        clear_iqxy,
        "#undef MAGNETIC",
        "#undef KERNEL_NAME",
    ]

    return iq, iqxy, imagnetic


def load_kernel_module(model_name):
    # type: (str) -> ModuleType
    """
    Return the kernel module named in *model_name*.

    If the name ends in *.py* then load it as a custom model using
    :func:`.custom.__init__.load_custom_kernel_module`, otherwise
    load it as a builtin from *sasmodels.models*.
    """
    if model_name.endswith('.py'):
        kernel_module = load_custom_kernel_module(model_name)
    else:
        try:
            from sasmodels import models
            __import__('sasmodels.models.'+model_name)
            kernel_module = getattr(models, model_name, None)
        except ImportError:
            # If the model isn't a built in model, try the plugin directory
            plugin_path = environ.get('SAS_MODELPATH', None)
            if plugin_path is not None:
                file_name = model_name.split(sep)[-1]
                model_name = plugin_path + sep + file_name + ".py"
                kernel_module = load_custom_kernel_module(model_name)
            else:
                raise
    return kernel_module


section_marker = re.compile(r'\A(?P<first>[%s])(?P=first)*\Z'
                            % re.escape(string.punctuation))
def _convert_section_titles_to_boldface(lines):
    # type: (Sequence[str]) -> Iterator[str]
    """
    Do the actual work of identifying and converting section headings.
    """
    prior = None
    for line in lines:
        if prior is None:
            prior = line
        elif section_marker.match(line):
            if len(line) >= len(prior):
                yield "".join(("**", prior, "**"))
                prior = None
            else:
                yield prior
                prior = line
        else:
            yield prior
            prior = line
    if prior is not None:
        yield prior


def convert_section_titles_to_boldface(s):
    # type: (str) -> str
    """
    Use explicit bold-face rather than section headings so that the table of
    contents is not polluted with section names from the model documentation.

    Sections are identified as the title line followed by a line of punctuation
    at least as long as the title line.
    """
    return "\n".join(_convert_section_titles_to_boldface(s.split('\n')))


def make_doc(model_info):
    # type: (ModelInfo) -> str
    """
    Return the documentation for the model.
    """
    Iq_units = "The returned value is scaled to units of |cm^-1| |sr^-1|, absolute scale."
    Sq_units = "The returned value is a dimensionless structure factor, $S(q)$."
    docs = model_info.docs if model_info.docs is not None else ""
    docs = convert_section_titles_to_boldface(docs)
    if model_info.structure_factor:
        pars = model_info.parameters.kernel_parameters
    else:
        pars = (model_info.parameters.common_parameters
                + model_info.parameters.kernel_parameters)
    partable = make_partable(pars)
    subst = dict(id=model_info.id.replace('_', '-'),
                 name=model_info.name,
                 title=model_info.title,
                 parameters=partable,
                 returns=Sq_units if model_info.structure_factor else Iq_units,
                 docs=docs)
    return DOC_HEADER % subst


# TODO: need a single source for rst_prolog; it is also in doc/rst_prolog
RST_PROLOG = r"""\
.. |Ang| unicode:: U+212B
.. |Ang^-1| replace:: |Ang|\ :sup:`-1`
.. |Ang^2| replace:: |Ang|\ :sup:`2`
.. |Ang^-2| replace:: |Ang|\ :sup:`-2`
.. |1e-6Ang^-2| replace:: 10\ :sup:`-6`\ |Ang|\ :sup:`-2`
.. |Ang^3| replace:: |Ang|\ :sup:`3`
.. |Ang^-3| replace:: |Ang|\ :sup:`-3`
.. |Ang^-4| replace:: |Ang|\ :sup:`-4`
.. |cm^-1| replace:: cm\ :sup:`-1`
.. |cm^2| replace:: cm\ :sup:`2`
.. |cm^-2| replace:: cm\ :sup:`-2`
.. |cm^3| replace:: cm\ :sup:`3`
.. |1e15cm^3| replace:: 10\ :sup:`15`\ cm\ :sup:`3`
.. |cm^-3| replace:: cm\ :sup:`-3`
.. |sr^-1| replace:: sr\ :sup:`-1`

.. |cdot| unicode:: U+00B7
.. |deg| unicode:: U+00B0
.. |g/cm^3| replace:: g\ |cdot|\ cm\ :sup:`-3`
.. |mg/m^2| replace:: mg\ |cdot|\ m\ :sup:`-2`
.. |fm^2| replace:: fm\ :sup:`2`
.. |Ang*cm^-1| replace:: |Ang|\ |cdot|\ cm\ :sup:`-1`
"""

# TODO: make a better fake reference role
RST_ROLES = """\
.. role:: ref

.. role:: numref

"""

def make_html(model_info):
    # type: (ModelInfo) -> str
    """
    Convert model docs directly to html.
    """
    from . import rst2html

    rst = make_doc(model_info)
    return rst2html.rst2html("".join((RST_ROLES, RST_PROLOG, rst)))

def view_html(model_name):
    # type: (str) -> None
    """
    Load the model definition and view its help.
    """
    from . import modelinfo
    kernel_module = load_kernel_module(model_name)
    info = modelinfo.make_model_info(kernel_module)
    view_html_from_info(info)

def view_html_from_info(info):
    # type: (ModelInfo) -> None
    """
    View the help for a loaded model definition.
    """
    from . import rst2html
    url = "file://"+dirname(info.filename)+"/"
    rst2html.view_html(make_html(info), url=url)

def demo_time():
    # type: () -> None
    """
    Show how long it takes to process a model.
    """
    import datetime
    from .modelinfo import make_model_info
    from .models import cylinder

    tic = datetime.datetime.now()
    make_source(make_model_info(cylinder))
    toc = (datetime.datetime.now() - tic).total_seconds()
    print("time: %g"%toc)


def main():
    # type: () -> None
    """
    Program which prints the source produced by the model.
    """
    from .modelinfo import make_model_info

    if len(sys.argv) <= 1:
        print("usage: python -m sasmodels.generate modelname")
    else:
        name = sys.argv[1]
        kernel_module = load_kernel_module(name)
        model_info = make_model_info(kernel_module)
        source = make_source(model_info)
        print(source['dll'])


if __name__ == "__main__":
    main()



r"""
Calculates the scattering from a randomly distributed, two-phase system based on
the Debye-Anderson-Brumberger (DAB) model for such systems. The two-phase system
is characterized by a single length scale, the correlation length, which is a
measure of the average spacing between regions of phase 1 and phase 2. **The
model also assumes smooth interfaces between the phases** and hence exhibits
Porod behavior $(I \sim q^{-4})$ at large $q$, $(qL \gg 1)$.

The DAB model is ostensibly a development of the earlier Debye-Bueche model.

Definition
----------

.. math::

    I(q) = \text{scale}\cdot\frac{L^3}{(1 + (q\cdot L)^2)^2} + \text{background}

where scale is

.. math:: \text{scale} = 8 \pi \phi (1-\phi) \Delta\rho^2

and the parameter $L$ is the correlation length.

For 2D data the scattering intensity is calculated in the same way as 1D,
where the $q$ vector is defined as

.. math:: q = \sqrt{q_x^2 + q_y^2}


References
----------

#. P Debye, H R Anderson, H Brumberger, *Scattering by an Inhomogeneous Solid.
   II. The Correlation Function and its Application*,
   *J. Appl. Phys.*, 28(6) (1957) 679
#. P Debye, A M Bueche, *Scattering by an Inhomogeneous Solid*,
   *J. Appl. Phys.*, 20 (1949) 518

Source
------

`dab.py <https://github.com/SasView/sasmodels/blob/master/sasmodels/models/dab.py>`_

Authorship and Verification
----------------------------

* **Author:**
* **Last Modified by:**
* **Last Reviewed by:** Steve King & Peter Parker **Date:** September 09, 2013
* **Source added by :** Steve King **Date:** March 25, 2019
"""

import numpy as np
from numpy import inf

name = "dab"
title = "DAB (Debye Anderson Brumberger) Model"
description = """\

F(q)= scale * L^3/(1 + (q*L)^2)^2

L: the correlation length

"""
category = "shape-independent"

#   ["name", "units", default, [lower, upper], "type", "description"],
parameters = [
    ["cor_length", "Ang", 50.0, [0, inf], "", "correlation length"],
    ]

source = ["dab.c"]

def random():
    """Return a random parameter set for the model."""
    pars = dict(
        scale=10**np.random.uniform(1, 4),
        cor_length=10**np.random.uniform(0.3, 3),
        #background=0,
    )
    pars['scale'] /= pars['cor_length']**3
    return pars



"""
Minimal example of calling a kernel for a specific set of q values.
"""

from numpy import logspace, sqrt
from matplotlib import pyplot as plt
from sasmodels.core import load_model
from sasmodels.direct_model import call_kernel, call_Fq

model = load_model('cylinder')
q = logspace(-3, -1, 200)
kernel = model.make_kernel([q])
pars = {'radius': 200, 'radius_pd': 0.1, 'scale': 2}
Iq = call_kernel(kernel, pars)
F, Fsq, Reff, V, Vratio = call_Fq(kernel, pars)
plt.loglog(q, Iq, label='2 I(q)')
plt.loglog(q, F**2/V, label='<F(q)>^2/V')
plt.loglog(q, Fsq/V, label='<F^2(q)>/V')
plt.xlabel('q (1/A)')
plt.ylabel('I(q) (1/cm)')
plt.title('Cylinder with radius 200.')
plt.legend()
plt.show()



r"""
Definition
----------

Calculates the form factor for a rectangular solid with a core-shell structure.
The thickness and the scattering length density of the shell or "rim" can be
different on each (pair) of faces. The three dimensions of the core of the
parallelepiped (strictly here a cuboid) may be given in *any* size order as
long as the particles are randomly oriented (i.e. take on all possible
orientations see notes on 2D below). To avoid multiple fit solutions,
especially with Monte-Carlo fit methods, it may be advisable to restrict their
ranges. There may be a number of closely similar "best fits", so some trial and
error, or fixing of some dimensions at expected values, may help.

The form factor is normalized by the particle volume $V$ such that

.. math::

    I(q) = \frac{\text{scale}}{V} \langle P(q,\alpha,\beta) \rangle
    + \text{background}

where $\langle \ldots \rangle$ is an average over all possible orientations
of the rectangular solid, and the usual $\Delta \rho^2 \ V^2$ term cannot be
pulled out of the form factor term due to the multiple slds in the model.

The core of the solid is defined by the dimensions $A$, $B$, $C$ here shown
such that $A < B < C$.

.. figure:: img/parallelepiped_geometry.jpg

   Core of the core shell parallelepiped with the corresponding definition
   of sides.


There are rectangular "slabs" of thickness $t_A$ that add to the $A$ dimension
(on the $BC$ faces). There are similar slabs on the $AC$ $(=t_B)$ and $AB$
$(=t_C)$ faces. The projection in the $AB$ plane is

.. figure:: img/core_shell_parallelepiped_projection.jpg

   AB cut through the core-shell parallelipiped showing the cross secion of
   four of the six shell slabs. As can be seen, this model leaves **"gaps"**
   at the corners of the solid.


The total volume of the solid is thus given as

.. math::

    V = ABC + 2t_ABC + 2t_BAC + 2t_CAB

The intensity calculated follows the :ref:`parallelepiped` model, with the
core-shell intensity being calculated as the square of the sum of the
amplitudes of the core and the slabs on the edges. The scattering amplitude is
computed for a particular orientation of the core-shell parallelepiped with
respect to the scattering vector and then averaged over all possible
orientations, where $\alpha$ is the angle between the $z$ axis and the $C$ axis
of the parallelepiped, and $\beta$ is the angle between the projection of the
particle in the $xy$ detector plane and the $y$ axis.

.. math::

    P(q)=\frac {\int_{0}^{\pi/2}\int_{0}^{\pi/2}F^2(q,\alpha,\beta) \ sin\alpha
    \ d\alpha \ d\beta} {\int_{0}^{\pi/2} \ sin\alpha \ d\alpha \ d\beta}

and

.. math::

    F(q,\alpha,\beta)
    &= (\rho_\text{core}-\rho_\text{solvent})
       S(Q_A, A) S(Q_B, B) S(Q_C, C) \\
    &+ (\rho_\text{A}-\rho_\text{solvent})
        \left[S(Q_A, A+2t_A) - S(Q_A, A)\right] S(Q_B, B) S(Q_C, C) \\
    &+ (\rho_\text{B}-\rho_\text{solvent})
        S(Q_A, A) \left[S(Q_B, B+2t_B) - S(Q_B, B)\right] S(Q_C, C) \\
    &+ (\rho_\text{C}-\rho_\text{solvent})
        S(Q_A, A) S(Q_B, B) \left[S(Q_C, C+2t_C) - S(Q_C, C)\right]

with

.. math::

    S(Q_X, L) = L \frac{\sin (\tfrac{1}{2} Q_X L)}{\tfrac{1}{2} Q_X L}

and

.. math::

    Q_A &= q \sin\alpha \sin\beta \\
    Q_B &= q \sin\alpha \cos\beta \\
    Q_C &= q \cos\alpha


where $\rho_\text{core}$, $\rho_\text{A}$, $\rho_\text{B}$ and $\rho_\text{C}$
are the scattering lengths of the parallelepiped core, and the rectangular
slabs of thickness $t_A$, $t_B$ and $t_C$, respectively. $\rho_\text{solvent}$
is the scattering length of the solvent.

.. note::

   the code actually implements two substitutions: $d(cos\alpha)$ is
   substituted for -$sin\alpha \ d\alpha$ (note that in the
   :ref:`parallelepiped` code this is explicitly implemented with
   $\sigma = cos\alpha$), and $\beta$ is set to $\beta = u \pi/2$ so that
   $du = \pi/2 \ d\beta$.  Thus both integrals go from 0 to 1 rather than 0
   to $\pi/2$.

FITTING NOTES
~~~~~~~~~~~~~

#. There are many parameters in this model. Hold as many fixed as possible with
   known values, or you will certainly end up at a solution that is unphysical.

#. The 2nd virial coefficient of the core_shell_parallelepiped is calculated
   based on the the averaged effective radius $(=\sqrt{(A+2t_A)(B+2t_B)/\pi})$
   and length $(C+2t_C)$ values, after appropriately sorting the three
   dimensions to give an oblate or prolate particle, to give an effective radius
   for $S(q)$ when $P(q) * S(q)$ is applied.

#. For 2d data the orientation of the particle is required, described using
   angles $\theta$, $\phi$ and $\Psi$ as in the diagrams below, where $\theta$
   and $\phi$ define the orientation of the director in the laboratory reference
   frame of the beam direction ($z$) and detector plane ($x-y$ plane), while
   the angle $\Psi$ is effectively the rotational angle around the particle
   $C$ axis. For $\theta = 0$ and $\phi = 0$, $\Psi = 0$ corresponds to the
   $B$ axis oriented parallel to the y-axis of the detector with $A$ along
   the x-axis. For other $\theta$, $\phi$ values, the order of rotations
   matters. In particular, the parallelepiped must first be rotated $\theta$
   degrees in the $x-z$ plane before rotating $\phi$ degrees around the $z$
   axis (in the $x-y$ plane). Applying orientational distribution to the
   particle orientation (i.e  `jitter` to one or more of these angles) can get
   more confusing as `jitter` is defined **NOT** with respect to the laboratory
   frame but the particle reference frame. It is thus highly recommended to
   read :ref:`orientation` for further details of the calculation and angular
   dispersions.

.. note:: For 2d, constraints must be applied during fitting to ensure that the
   order of sides chosen is not altered, and hence that the correct definition
   of angles is preserved. For the default choice shown here, that means
   ensuring that the inequality $A < B < C$ is not violated,  The calculation
   will not report an error, but the results may be not correct.

.. figure:: img/parallelepiped_angle_definition.png

    Definition of the angles for oriented core-shell parallelepipeds.
    Note that rotation $\theta$, initially in the $x-z$ plane, is carried
    out first, then rotation $\phi$ about the $z$ axis, finally rotation
    $\Psi$ is now around the $C$ axis of the particle. The neutron or X-ray
    beam is along the $z$ axis and the detecotr defines the $x-y$ plane.

.. figure:: img/parallelepiped_angle_projection.png

    Examples of the angles for oriented core-shell parallelepipeds against the
    detector plane.


Validation
----------

Cross-checked against hollow rectangular prism and rectangular prism for equal
thickness overlapping sides, and by Monte Carlo sampling of points within the
shape for non-uniform, non-overlapping sides.


References
----------

#. P Mittelbach and G Porod, *Acta Physica Austriaca*, 14 (1961) 185-211
   Equations (1), (13-14). (in German)
#. D Singh (2009). *Small angle scattering studies of self assembly in
   lipid mixtures*, Johns Hopkins University Thesis (2009) 223-225. `Available
   from Proquest <http://search.proquest.com/docview/304915826>`_
#.  L. Onsager, *Ann. New York Acad. Sci.*, 51 (1949) 627-659

Authorship and Verification
----------------------------

* **Author:** NIST IGOR/DANSE **Date:** pre 2010
* **Converted to sasmodels by:** Miguel Gonzalez **Date:** February 26, 2016
* **Last Modified by:** Paul Kienzle **Date:** October 17, 2017
* **Last Reviewed by:** Paul Butler **Date:** May 24, 2018 - documentation
  updated
"""

import numpy as np
from numpy import inf

name = "core_shell_parallelepiped"
title = "Rectangular solid with a core-shell structure."
description = """
     P(q)=
"""
category = "shape:parallelepiped"

#             ["name", "units", default, [lower, upper], "type","description"],
parameters = [["sld_core", "1e-6/Ang^2", 1, [-inf, inf], "sld",
               "Parallelepiped core scattering length density"],
              ["sld_a", "1e-6/Ang^2", 2, [-inf, inf], "sld",
               "Parallelepiped A rim scattering length density"],
              ["sld_b", "1e-6/Ang^2", 4, [-inf, inf], "sld",
               "Parallelepiped B rim scattering length density"],
              ["sld_c", "1e-6/Ang^2", 2, [-inf, inf], "sld",
               "Parallelepiped C rim scattering length density"],
              ["sld_solvent", "1e-6/Ang^2", 6, [-inf, inf], "sld",
               "Solvent scattering length density"],
              ["length_a", "Ang", 35, [0, inf], "volume",
               "Shorter side of the parallelepiped"],
              ["length_b", "Ang", 75, [0, inf], "volume",
               "Second side of the parallelepiped"],
              ["length_c", "Ang", 400, [0, inf], "volume",
               "Larger side of the parallelepiped"],
              ["thick_rim_a", "Ang", 10, [0, inf], "volume",
               "Thickness of A rim"],
              ["thick_rim_b", "Ang", 10, [0, inf], "volume",
               "Thickness of B rim"],
              ["thick_rim_c", "Ang", 10, [0, inf], "volume",
               "Thickness of C rim"],
              ["theta", "degrees", 0, [-360, 360], "orientation",
               "c axis to beam angle"],
              ["phi", "degrees", 0, [-360, 360], "orientation",
               "rotation about beam"],
              ["psi", "degrees", 0, [-360, 360], "orientation",
               "rotation about c axis"],
             ]

source = ["lib/gauss76.c", "core_shell_parallelepiped.c"]
have_Fq = True
radius_effective_modes = [
    "equivalent cylinder excluded volume",
    "equivalent volume sphere",
    "half outer length_a", "half outer length_b", "half outer length_c",
    "equivalent circular cross-section",
    "half outer ab diagonal", "half outer diagonal",
    ]

def random():
    """Return a random parameter set for the model."""
    outer = 10**np.random.uniform(1, 4.7, size=3)
    thick = np.random.beta(0.5, 0.5, size=3)*(outer-2) + 1
    length = outer - thick
    pars = dict(
        length_a=length[0],
        length_b=length[1],
        length_c=length[2],
        thick_rim_a=thick[0],
        thick_rim_b=thick[1],
        thick_rim_c=thick[2],
    )
    return pars

# rkh 7/4/17 add random unit test for 2d, note make all params different,
# 2d values not tested against other codes or models
if 0:  # pak: model rewrite; need to update tests
    qx, qy = 0.2 * np.cos(np.pi/6.), 0.2 * np.sin(np.pi/6.)
    tests = [[{}, 0.2, 0.533149288477],
             [{}, [0.2], [0.533149288477]],
             [{'theta':10.0, 'phi':20.0}, (qx, qy), 0.0853299803222],
             [{'theta':10.0, 'phi':20.0}, [(qx, qy)], [0.0853299803222]],
            ]
    del qx, qy  # not necessary to delete, but cleaner



r"""
Definition
----------
Calculates the scattering from a fractal structure with a primary building
block of core-shell spheres, as opposed to just homogeneous spheres in the
fractal model. It is an extension of the well known Teixeira\ [#Teixeira1988]_
fractal model replacing the $P(q)$ of a solid sphere with that of a core-shell
sphere. This model could find use for aggregates of coated particles, or
aggregates of vesicles for example.

.. math::

    I(q) = P(q)S(q) + \text{background}

Where $P(q)$ is the core-shell form factor and $S(q)$ is the
Teixeira\ [#Teixeira1988]_ fractal structure factor both of which are given
again below:

.. math::

    P(q) &= \frac{\phi}{V_s}\left[3V_c(\rho_c-\rho_s)
    \frac{\sin(qr_c)-qr_c\cos(qr_c)}{(qr_c)^3}+
    3V_s(\rho_s-\rho_{solv})
    \frac{\sin(qr_s)-qr_s\cos(qr_s)}{(qr_s)^3}\right]^2 \\
    S(q) &= 1 + \frac{D_f\ \Gamma\!(D_f-1)}{[1+1/(q\xi)^2]^{(D_f-1)/2}}
    \frac{\sin[(D_f-1)\tan^{-1}(q\xi)]}{(qr_s)^{D_f}}

where $\phi$ is the volume fraction of particles, $V_s$ is the volume of the
whole particle, $V_c$ is the volume of the core, $\rho_c$, $\rho_s$, and
$\rho_{solv}$ are the scattering length densities of the core, shell, and
solvent respectively, $r_c$ and $r_s$ are the radius of the core and the
radius of the whole particle respectively, $D_f$ is the fractal dimension,
and $\xi$ the correlation length.

Polydispersity of radius and thickness are also provided for.

This model does not allow for anisotropy and thus the 2D scattering intensity
is calculated in the same way as 1D, where the $q$ vector is defined as

.. math::

    q = \sqrt{q_x^2 + q_y^2}

Our model is derived from the form factor calculations implemented in IGOR
macros by the NIST Center for Neutron Research\ [#Kline2006]_

References
----------

.. [#Teixeira1988] J Teixeira, *J. Appl. Cryst.*, 21 (1988) 781-785
.. [#Kline2006]  S R Kline, *J Appl. Cryst.*, 39 (2006) 895

Authorship and Verification
----------------------------

* **Author:** NIST IGOR/DANSE **Date:** pre 2010
* **Last Modified by:** Paul Butler and Paul Kienzle **Date:** November 27, 2016
* **Last Reviewed by:** Paul Butler and Paul Kienzle **Date:** November 27, 2016
"""

import numpy as np
from numpy import inf

name = "fractal_core_shell"
title = "Scattering from a fractal structure formed from core shell spheres"
description = """\
    Model for fractal aggregates of core-shell primary particles. It is based on
    the Teixeira model for the S(q) of a fractal * P(q) for a core-shell sphere

    radius =  the radius of the core
    thickness = thickness of the shell
    thick_layer = thickness of a layer
    sld_core = the SLD of the core
    sld_shell = the SLD of the shell
    sld_solvent = the SLD of the solvent
    volfraction = volume fraction of core-shell particles
    fractal_dim = fractal dimension
    cor_length = correlation length of the fractal like aggretates
    """
category = "shape-independent"

# pylint: disable=bad-whitespace, line-too-long
#   ["name", "units", default, [lower, upper], "type","description"],
parameters = [
    ["radius",      "Ang",        60.0, [0.0, inf],  "volume", "Sphere core radius"],
    ["thickness",   "Ang",        10.0, [0.0, inf],  "volume", "Sphere shell thickness"],
    ["sld_core",    "1e-6/Ang^2", 1.0,  [-inf, inf], "sld",    "Sphere core scattering length density"],
    ["sld_shell",   "1e-6/Ang^2", 2.0,  [-inf, inf], "sld",    "Sphere shell scattering length density"],
    ["sld_solvent", "1e-6/Ang^2", 3.0,  [-inf, inf], "sld",    "Solvent scattering length density"],
    ["volfraction", "",           0.05,  [0.0, inf],  "",       "Volume fraction of building block spheres"],
    ["fractal_dim",    "",        2.0,  [0.0, 6.0],  "",       "Fractal dimension"],
    ["cor_length",  "Ang",      100.0,  [0.0, inf],  "",       "Correlation length of fractal-like aggregates"],
]
# pylint: enable=bad-whitespace, line-too-long

source = ["lib/sas_3j1x_x.c", "lib/sas_gamma.c", "lib/core_shell.c",
          "lib/fractal_sq.c", "fractal_core_shell.c"]

def random():
    """Return a random parameter set for the model."""
    outer_radius = 10**np.random.uniform(0.7, 4)
    # Use a distribution with a preference for thin shell or thin core
    # Avoid core,shell radii < 1
    thickness = np.random.beta(0.5, 0.5)*(outer_radius-2) + 1
    radius = outer_radius - thickness
    cor_length = 10**np.random.uniform(0.7, 2)*outer_radius
    volfraction = 10**np.random.uniform(-3, -1)
    fractal_dim = 2*np.random.beta(3, 4) + 1
    pars = dict(
        #background=0, sld_block=1, sld_solvent=0,
        volfraction=volfraction,
        radius=radius,
        cor_length=cor_length,
        fractal_dim=fractal_dim,
    )
    return pars

#tests = [[{'radius': 20.0, 'thickness': 10.0}, 'ER', 30.0],
tests = [
    # At some point the SasView 3.x test result was deemed incorrect.  The
    # following tests were verified against NIST IGOR macros ver 7.850.
    # NOTE: NIST macros do only provide for a polydisperse core (no option
    # for a poly shell or for a monodisperse core.  The results seemed
    # extremely sensitive to the core PD, varying non monotonically all
    # the way to a PD of 1e-6. From 1e-6 to 1e-9 no changes in the
    # results were observed and the values below were taken using PD=1e-9.
    # Non-monotonically = I(0.001)=188 to 140 to 177 back to 160 etc.
    [{'radius': 20.0,
      'thickness': 5.0,
      'sld_core': 3.5,
      'sld_shell': 1.0,
      'sld_solvent': 6.35,
      'volfraction': 0.05,
      'background': 0.0},
     [0.001, 0.00291, 0.0107944, 0.029923, 0.100726, 0.476304],
     [177.146, 165.151, 84.1596, 20.1466, 1.40906, 0.00622666]
    ]
]



from __future__ import with_statement
import os.path

MODULE_TEMPLATE = """.. Autogenerated by genmods.py

******************************************************************************
%(name)s
******************************************************************************

:mod:`%(package)s.%(module)s`
==============================================================================

.. automodule:: %(package)s.%(module)s
   :members:
   :undoc-members:
   :inherited-members:
   :show-inheritance:

"""

INDEX_TEMPLATE = """.. Autogenerated by genmods.py

.. _api-index:

##############################################################################
   %(package_name)s
##############################################################################

.. only:: html

   :Release: |version|
   :Date: |today|

.. toctree::
   :titlesonly:
   :numbered: 1
   :maxdepth: 2

   %(rsts)s
"""


def genfiles(package, package_name, modules, dir='api'):

    if not os.path.exists(dir):
        os.makedirs(dir)

    for module, name in modules:
        with open(os.path.join(dir, module+'.rst'), 'w') as f:
            f.write(MODULE_TEMPLATE%locals())

    rsts = "\n   ".join(module+'.rst' for module, name in modules)
    with open(os.path.join(dir, 'index.rst'), 'w') as f:
        f.write(INDEX_TEMPLATE%locals())


modules = [
    ('__init__', 'Sasmodels package'),
    #('alignment', 'GPU data alignment [unused]'),
    ('bumps_model', 'Bumps interface'),
    ('compare', 'Compare models on different compute engines'),
    ('compare_many', 'Batch compare models on different compute engines'),
    ('conversion_table', 'Model conversion table'),
    ('convert', 'Sasview to sasmodel converter'),
    ('custom.__init__', 'Load custom kernel modules'),
    ('core', 'Model access'),
    ('data', 'Data layout and plotting routines'),
    ('details', 'Parameter packing for kernel calls'),
    ('direct_model', 'Simple interface'),
    ('exception', 'Annotate exceptions'),
    ('generate', 'Model parser'),
    ('gengauss', 'Generate Gauss-Legendre integration points'),
    ('guyou', 'Guyou map projection'),
    ('jitter', 'Orientation explorer'),
    ('kernel', 'Evaluator type definitions'),
    ('kernelcl', 'OpenCL model evaluator'),
    ('kernelcuda', 'CUDA model evaluator'),
    ('kerneldll', 'Ctypes model evaluator'),
    ('kernelpy', 'Python model evaluator'),
    ('list_pars', 'Identify all parameters in all models'),
    ('mixture', 'Mixture model evaluator'),
    # Docs for models are generated by Makefile + genmodel.py, and don't
    # need to be managed by autodoc.
    #('models.__init__', 'Builtin models'),
    ('model_test', 'Unit test support'),
    ('modelinfo', 'Parameter and model definitions'),
    ('multiscat', 'Multiple scattering support'),
    ('product', 'Product model evaluator'),
    ('resolution', '1-D resolution functions'),
    ('resolution2d', '2-D resolution functions'),
    ('rst2html', 'Convert doc strings the web pages'),
    ('sasview_model', 'Sasview interface'),
    ('sesans', 'SESANS calculation routines'),
    ('special', 'Special functions library'),
    ('weights', 'Distribution functions'),
]
package = 'sasmodels'
package_name = 'Reference'
genfiles(package, package_name, modules)



# Note: model title and parameter table are inserted automatically
r"""
Calculates the interparticle structure factor for monodisperse
spherical particles interacting through hard sphere (excluded volume)
interactions. This $S(q)$ may also be a reasonable approximation for
other particle shapes that freely rotate (but see the note below),
and for moderately polydisperse systems.

.. note::

   This routine is intended for uncharged particles! For charged
   particles try using the :ref:`hayter-msa` $S(q)$ instead.

.. note::

   Earlier versions of SasView did not incorporate the so-called
   $\beta(q)$ ("beta") correction [1] for polydispersity and non-sphericity.
   This is only available in SasView versions 5.0 and higher.

radius_effective is the effective hard sphere radius.
volfraction is the volume fraction occupied by the spheres.

In SasView the effective radius may be calculated from the parameters
used in the form factor $P(q)$ that this $S(q)$ is combined with.

For numerical stability the computation uses a Taylor series expansion
at very small $qR$, but there may be a very minor glitch at the
transition point in some circumstances.

This S(q) uses the Percus-Yevick closure relationship [2] where the
interparticle potential $U(r)$ is

.. math::

    U(r) = \begin{cases}
    \infty & r < 2R \\
    0 & r \geq 2R
    \end{cases}

where $r$ is the distance from the center of a sphere of a radius $R$.

For a 2D plot, the wave transfer is defined as

.. math::

    q = \sqrt{q_x^2 + q_y^2}


References
----------

#.  M Kotlarchyk & S-H Chen, *J. Chem. Phys.*, 79 (1983) 2461-2469

#.  J K Percus, J Yevick, *J. Phys. Rev.*, 110, (1958) 1

Authorship and Verification
----------------------------

* **Author:**
* **Last Modified by:**
* **Last Reviewed by:**
"""

import numpy as np
from numpy import inf

name = "hardsphere"
title = "Hard sphere structure factor, with Percus-Yevick closure"
description = """\
    [Hard sphere structure factor, with Percus-Yevick closure]
        Interparticle S(Q) for random, non-interacting spheres.
    May be a reasonable approximation for other particle shapes
    that freely rotate, and for moderately polydisperse systems
    . The "beta(q)" correction is available in versions 4.2.2
    and higher.
    radius_effective is the hard sphere radius
    volfraction is the volume fraction occupied by the spheres.
"""
category = "structure-factor"
structure_factor = True
single = False # TODO: check

#             ["name", "units", default, [lower, upper], "type","description"],
parameters = [["radius_effective", "Ang", 50.0, [0, inf], "",
               "effective radius of hard sphere"],
              ["volfraction", "", 0.2, [0, 0.74], "",
               "volume fraction of hard spheres"],
             ]

source = ["hardsphere.c"]

def random():
    """Return a random parameter set for the model."""
    pars = dict(
        scale=1, background=0,
        radius_effective=10**np.random.uniform(1, 4),
        volfraction=10**np.random.uniform(-2, 0),  # high volume fraction
    )
    return pars

# Q=0.001 is in the Taylor series, low Q part, so add Q=0.1,
# assuming double precision sasview is correct
tests = [
    [{'scale': 1.0, 'background' : 0.0, 'radius_effective' : 50.0,
      'volfraction' : 0.2},
     [0.001, 0.1], [0.209128, 0.930587]],
]
# ADDED by: RKH  ON: 16Mar2016  using equations from FISH as better than
# orig sasview, see notes above. Added Taylor expansions at small Q.



"""
Class interface to the model calculator.

Calling a model is somewhat non-trivial since the functions called depend
on the data type.  For 1D data the *Iq* kernel needs to be called, for
2D data the *Iqxy* kernel needs to be called, and for SESANS data the
*Iq* kernel needs to be called followed by a Hankel transform.  Before
the kernel is called an appropriate *q* calculation vector needs to be
constructed.  This is not the simple *q* vector where you have measured
the data since the resolution calculation will require values beyond the
range of the measured data.  After the calculation the resolution calculator
must be called to return the predicted value for each measured data point.

:class:`DirectModel` is a callable object that takes *parameter=value*
keyword arguments and returns the appropriate theory values for the data.

:class:`DataMixin` does the real work of interpreting the data and calling
the model calculator.  This is used by :class:`DirectModel`, which uses
direct parameter values and by :class:`.bumps_model.Experiment` which wraps
the parameter values in boxes so that the user can set fitting ranges, etc.
on the individual parameters and send the model to the Bumps optimizers.
"""
from __future__ import print_function, division

import os

import numpy as np  # type: ignore

# TODO: fix sesans module
from . import sesans  # type: ignore
from . import weights
from . import resolution
from . import resolution2d
from .details import make_kernel_args, dispersion_mesh
from .product import RADIUS_MODE_ID

# pylint: disable=unused-import
from typing import Optional, Dict, Tuple, List, Callable
from collections import OrderedDict
from .data import Data
from .kernel import Kernel, KernelModel
from .modelinfo import Parameter, ParameterSet, ModelInfo
# pylint: enable=unused-import

def call_kernel(calculator, pars, cutoff=0., mono=False):
    # type: (Kernel, ParameterSet, float, bool) -> np.ndarray
    """
    Call *kernel* returned from *model.make_kernel* with parameters *pars*.

    *cutoff* is the limiting value for the product of dispersion weights used
    to perform the multidimensional dispersion calculation more quickly at a
    slight cost to accuracy. The default value of *cutoff=0* integrates over
    the entire dispersion cube.  Using *cutoff=1e-5* can be 50% faster, but
    with an error of about 1%, which is usually less than the measurement
    uncertainty.

    *mono* is True if polydispersity should be set to none on all parameters.
    """
    mesh = get_mesh(calculator.info, pars, dim=calculator.dim, mono=mono)
    #print("in call_kernel: pars:", list(zip(*mesh))[0])
    call_details, values, is_magnetic = make_kernel_args(calculator, mesh)
    #print("in call_kernel: values:", values)
    return calculator(call_details, values, cutoff, is_magnetic)

def call_Fq(calculator, pars, cutoff=0., mono=False):
    # type: (Kernel, ParameterSet, float, bool) -> np.ndarray
    """
    Like :func:`call_kernel`, but returning F, F^2, R_eff, V_shell, V_form/V_shell.

    For solid objects V_shell is equal to V_form and the volume ratio is 1.

    Use parameter *radius_effective_mode* to select the effective radius
    calculation to use amongst the *radius_effective_modes* list given in the
    model.
    """
    R_eff_type = int(pars.pop(RADIUS_MODE_ID, 1.0))
    mesh = get_mesh(calculator.info, pars, dim=calculator.dim, mono=mono)
    #print("in call_Fq: pars", list(zip(*mesh))[0])
    call_details, values, is_magnetic = make_kernel_args(calculator, mesh)
    #print("in call_Fq: values:", values)
    return calculator.Fq(call_details, values, cutoff, is_magnetic, R_eff_type)

def call_profile(model_info, pars=None):
    # type: (ModelInfo, ParameterSet) -> Tuple[np.ndarray, np.ndarray, Tuple[str, str]]
    """
    Returns the profile *x, y, (xlabel, ylabel)* representing the model.
    """
    if pars is None:
        pars = {}
    args = {}
    for p in model_info.parameters.kernel_parameters:
        if p.length > 1:
            value = np.array([pars.get(p.id+str(j), p.default)
                              for j in range(1, p.length+1)])
        else:
            value = pars.get(p.id, p.default)
        args[p.id] = value
    x, y = model_info.profile(**args)
    return x, y, model_info.profile_axes

def get_mesh(model_info, values, dim='1d', mono=False):
    # type: (ModelInfo, Dict[str, float], str, bool) -> List[Tuple[float, np.ndarray, np.ndarray]]
    """
    Retrieve the dispersity mesh described by the parameter set.

    Returns a list of *(value, dispersity, weights)* with one tuple for each
    parameter in the model call parameters.  Inactive parameters return the
    default value with a weight of 1.0.
    """
    parameters = model_info.parameters
    if mono:
        active = lambda name: False
    elif dim == '1d':
        active = lambda name: name in parameters.pd_1d
    elif dim == '2d':
        active = lambda name: name in parameters.pd_2d
    else:
        active = lambda name: True

    #print("in get_mesh: pars:",[p.id for p in parameters.call_parameters])
    values = values.copy()
    mesh = [_pop_par_weights(p, values, active(p.name))
            for p in parameters.call_parameters]

    if values:
        raise TypeError(f"Unused parameters in call: {', '.join(values.keys())}")

    return mesh


def _pop_par_weights(parameter, values, active=True):
    # type: (Parameter, Dict[str, float], bool) -> Tuple[float, np.ndarray, np.ndarray]
    """
    Generate the distribution for parameter *name* given the parameter values
    in *pars*.

    Uses "name", "name_pd", "name_pd_type", "name_pd_n", "name_pd_sigma"
    from the *pars* dictionary for parameter value and parameter dispersion.
    """
    value = float(values.pop(parameter.name, parameter.default))
    if parameter.polydisperse:
        npts = values.pop(parameter.name+'_pd_n', 0)
        width = values.pop(parameter.name+'_pd', 0.0)
        nsigma = values.pop(parameter.name+'_pd_nsigma', 3.0)
        distribution = values.pop(parameter.name+'_pd_type', 'gaussian')
        relative = parameter.relative_pd
        if npts == 0 or width == 0.0 or not active:
            # Note: orientation parameters have the viewing angle as the parameter
            # value and the jitter in the distribution, so be sure to set the
            # empty pd for orientation parameters to 0.
            pd = [value if relative else 0.0], [1.0]
        else:
            limits = parameter.limits
            pd = weights.get_weights(distribution, npts, width, nsigma,
                                    value, limits, relative)
    else:
        pd = [value], [1.0]
    return value, pd[0], pd[1]


def _make_sesans_transform(data):
    # Pre-compute the Hankel matrix (H)
    SElength, SEunits = data.x, data._xunit
    wavelength, wunits = data.source.wavelength, data.source.wavelength_unit
    theta_max, theta_units = data.sample.zacceptance
    if SEunits != "A" or wunits != "A" or theta_units != "radians":
        try:
            from sasdata.data_util.nxsunit import Converter
        except ImportError as ie:
            raise ImportError(f"{ie.name} is not available. Add sasdata to the python path.")

        SElength = Converter("A")(SElength, units=SEunits)
        wavelength = Converter("A")(wavelength, units=wunits)
        theta_max = Converter("radian")(theta_max, units=theta_units)

    Rmax = 10000000
    zaccept = 2 * np.pi / np.max(wavelength) * np.sin(theta_max)
    hankel = sesans.SesansTransform(data.x, SElength, wavelength, zaccept, Rmax)
    return hankel


class DataMixin(object):
    """
    DataMixin captures the common aspects of evaluating a SAS model for a
    particular data set, including calculating Iq and evaluating the
    resolution function.  It is used in particular by :class:`DirectModel`,
    which evaluates a SAS model parameters as key word arguments to the
    calculator method, and by :class:`.bumps_model.Experiment`, which wraps the
    model and data for use with the Bumps fitting engine.  It is not
    currently used by :class:`.sasview_model.SasviewModel` since this will
    require a number of changes to SasView before we can do it.

    *_interpret_data* initializes the data structures necessary
    to manage the calculations.  This sets attributes in the child class
    such as *data_type* and *resolution*.

    *_calc_theory* evaluates the model at the given control values.

    *_set_data* sets the intensity data in the data object,
    possibly with random noise added.  This is useful for simulating a
    dataset with the results from *_calc_theory*.
    """
    def _interpret_data(self, data: Data, model: KernelModel) -> None:
        # not type: (Data, KernelModel) -> None
        # pylint: disable=attribute-defined-outside-init

        self._data = data
        self._model = model

        # interpret data
        if getattr(data, 'isSesans', False):
            self.data_type = 'sesans'
        elif hasattr(data, 'qx_data'):
            self.data_type = 'Iqxy'
        elif getattr(data, 'oriented', False):
            self.data_type = 'Iq-oriented'
        else:
            self.data_type = 'Iq'

        if self.data_type == 'sesans':
            res = _make_sesans_transform(data)
            index = slice(None, None)
            if data.y is not None:
                Iq, dIq = data.y, data.dy
            else:
                Iq, dIq = None, None
        elif self.data_type == 'Iqxy':
            #if not model.info.parameters.has_2d:
            #    raise ValueError("not 2D without orientation or magnetic parameters")
            q = np.sqrt(data.qx_data**2 + data.qy_data**2)
            qmin = getattr(data, 'qmin', 1e-16)
            qmax = getattr(data, 'qmax', np.inf)
            accuracy = getattr(data, 'accuracy', 'Low')
            index = (data.mask == 0) & (q >= qmin) & (q <= qmax)
            if data.data is not None:
                index &= ~np.isnan(data.data)
                Iq = data.data[index]
                dIq = data.err_data[index]
            else:
                Iq, dIq = None, None
            res = resolution2d.Pinhole2D(data=data, index=index,
                                         nsigma=3.0, accuracy=accuracy)
        elif self.data_type == 'Iq':
            index = (data.x >= data.qmin) & (data.x <= data.qmax)
            mask = getattr(data, 'mask', None)
            if mask is not None:
                index &= (mask == 0)
            if data.y is not None:
                index &= ~np.isnan(data.y)
                Iq = data.y[index]
                dIq = data.dy[index]
            else:
                Iq, dIq = None, None
            if getattr(data, 'dx', None) is not None:
                q, dq = data.x[index], data.dx[index]
                if (dq > 0).any():
                    res = resolution.Pinhole1D(q, dq)
                else:
                    res = resolution.Perfect1D(q)
            elif (getattr(data, 'dxl', None) is not None
                  or getattr(data, 'dxw', None) is not None):
                res = resolution.Slit1D(
                    data.x[index],
                    q_length=None if data.dxl is None else data.dxl[index],
                    q_width=None if data.dxw is None else data.dxw[index])
            else:
                res = resolution.Perfect1D(data.x[index])
        elif self.data_type == 'Iq-oriented':
            index = (data.x >= data.qmin) & (data.x <= data.qmax)
            if data.y is not None:
                index &= ~np.isnan(data.y)
                Iq = data.y[index]
                dIq = data.dy[index]
            else:
                Iq, dIq = None, None
            if (getattr(data, 'dxl', None) is None
                    or getattr(data, 'dxw', None) is None):
                raise ValueError("oriented sample with 1D data needs slit resolution")

            res = resolution2d.Slit2D(
                data.x[index],
                qx_width=data.dxw[index],
                qy_width=data.dxl[index])
        else:
            raise ValueError("Unknown data type") # never gets here

        # Remember function inputs so we can delay loading the function and
        # so we can save/restore state
        self._kernel = None
        self.Iq, self.dIq, self.index = Iq, dIq, index
        self.resolution = res
        self.results = None  # type: Optional[Callable[[], OrderedDict]]

    def _set_data(self, Iq, noise=None):
        # type: (np.ndarray, Optional[float]) -> None
        # pylint: disable=attribute-defined-outside-init
        if noise is not None:
            self.dIq = Iq*noise*0.01
        dy = self.dIq
        y = Iq + np.random.randn(*dy.shape) * dy
        self.Iq = y
        if self.data_type in ('Iq', 'Iq-oriented'):
            if self._data.y is None:
                self._data.y = np.empty(len(self._data.x), 'd')
            if self._data.dy is None:
                self._data.dy = np.empty(len(self._data.x), 'd')
            self._data.dy[self.index] = dy
            self._data.y[self.index] = y
        elif self.data_type == 'Iqxy':
            if self._data.data is None:
                self._data.data = np.empty_like(self._data.qx_data, 'd')
            if self._data.err_data is None:
                self._data.err_data = np.empty_like(self._data.qx_data, 'd')
            self._data.data[self.index] = y
            self._data.err_data[self.index] = dy
        elif self.data_type == 'sesans':
            if self._data.y is None:
                self._data.y = np.empty(len(self._data.x), 'd')
            self._data.y[self.index] = y
        else:
            raise ValueError("Unknown model")

    def _calc_theory(self, pars, cutoff=0.0):
        # type: (ParameterSet, float) -> np.ndarray
        if self._kernel is None:
            # TODO: change interfaces so that resolution returns kernel inputs
            # Maybe have resolution always return a tuple, or maybe have
            # make_kernel accept either an ndarray or a pair of ndarrays.
            kernel_inputs = self.resolution.q_calc
            if isinstance(kernel_inputs, np.ndarray):
                kernel_inputs = (kernel_inputs,)
            self._kernel = self._model.make_kernel(kernel_inputs)

        # Need to pull background out of resolution for multiple scattering
        default_background = self._model.info.parameters.common_parameters[1].default
        background = (
            pars.get('background', default_background)
            if self.data_type != 'sesans' else 0.)
        pars = pars.copy()
        pars['background'] = 0.

        Iq_calc = call_kernel(self._kernel, pars, cutoff=cutoff)
        self.results = getattr(self._kernel, 'results', None)
        # Storing the calculated Iq values so that they can be plotted.
        # Only applies to oriented USANS data for now.
        # TODO: extend plotting of calculate Iq to other measurement types
        # TODO: refactor so we don't store the result in the model
        self.Iq_calc = Iq_calc
        result = self.resolution.apply(Iq_calc)
        if hasattr(self.resolution, 'nx'):
            self.Iq_calc = (
                self.resolution.qx_calc, self.resolution.qy_calc,
                np.reshape(Iq_calc, (self.resolution.ny, self.resolution.nx))
            )
        return result + background


class DirectModel(DataMixin):
    """
    Create a calculator object for a model.

    *data* is 1D SAS, 2D SAS or SESANS data

    *model* is a model calculator return from :func:`.core.load_model`

    *cutoff* is the polydispersity weight cutoff.
    """
    def __init__(self, data: Data, model: KernelModel, cutoff: float=1e-5) -> None:
        # not type: (Data, KernelModel, float) -> None
        self.model = model
        self.cutoff = cutoff
        # Note: _interpret_data defines the model attributes
        self._interpret_data(data, model)

    def __call__(self, **pars):
        # type: (**float) -> np.ndarray
        return self._calc_theory(pars, cutoff=self.cutoff)

    def simulate_data(self, noise=None, **pars):
        # type: (Optional[float], **float) -> None
        """
        Generate simulated data for the model.
        """
        Iq = self.__call__(**pars)
        self._set_data(Iq, noise=noise)

    def profile(self, **pars):
        # type: (**float) -> None
        """
        Generate a plottable profile.
        """
        return call_profile(self.model.info, pars)

def test_reparameterize():
    # type: () -> None
    """Check simple reparameterized models will load and build"""
    from numpy import inf, pi
    from numpy.linalg import norm
    from .data import empty_data1D
    from .core import load_model_info, build_model, reparameterize
    parameters = [
        ["volume", "Ang^3", 1e5, [0, inf], "volume", "ellipsoid volume"],
        ["eccentricity", "", 1, [0, inf], "volume", "polar:equatorial radius"],
    ]
    translation = """
        Re = cbrt(volume/eccentricity/M_4PI_3)
        radius_polar = eccentricity*Re
        radius_equatorial = Re  # python style comments allowed
        """
    info = reparameterize('ellipsoid', parameters, translation)

    # Check translated parameters
    actual = [p.name for p in info.parameters.kernel_parameters]
    target = ["sld", "sld_solvent", "volume", "eccentricity", "theta", "phi"]
    assert target == actual, "%s != %s"%(target, actual)

    data = empty_data1D([0.01, 0.1])
    Rp, Re = 30, 50

    base_model = build_model(load_model_info('ellipsoid'), dtype='d')
    base_calculator = DirectModel(data, base_model)
    base_Iq = base_calculator(radius_polar=Rp, radius_equatorial=Re)

    derived_model = build_model(info, dtype='d')
    derived_calculator = DirectModel(data, derived_model)
    derived_Iq = derived_calculator(volume=4*pi/3*Rp*Re**2, eccentricity=Rp/Re)
    assert norm((base_Iq - derived_Iq)/base_Iq) < 1e-13

    # clean up the dll that was created in the cache as part of the test
    try:
        os.remove(derived_model.dllpath)
    except Exception:
        pass

    # Again using insert_after.
    insert_after = {
        '': 'eccentricity',
        'radius_equatorial': 'volume',
        }
    info = reparameterize('ellipsoid', parameters, translation,
                          insert_after=insert_after)
    # Check translated parameters
    actual = [p.name for p in info.parameters.kernel_parameters]
    target = ["eccentricity", "sld", "sld_solvent", "volume", "theta", "phi"]
    assert target == actual, "%s != %s"%(target, actual)

    # Check calculation.
    derived_model = build_model(info, dtype='d')
    derived_calculator = DirectModel(data, derived_model)
    derived_Iq = derived_calculator(volume=4*pi/3*Rp*Re**2, eccentricity=Rp/Re)
    assert norm((base_Iq - derived_Iq)/base_Iq) < 1e-13

    # clean up the dll that was created in the cache as part of the test
    try:
        os.remove(derived_model.dllpath)
    except Exception:
        pass

def _direct_calculate(model, data, pars):
    from .core import load_model_info, build_model
    model_info = load_model_info(model)
    kernel = build_model(model_info)
    calculator = DirectModel(data, kernel)
    return calculator(**pars)

def Iq(model, q, dq=None, ql=None, qw=None, **pars):
    """
    Compute I(q) for *model*. Resolution is *dq* for pinhole or *ql* and *qw*
    for slit geometry. Use 0 or None for infinite slits.

    Model is the name of a builtin or custom model, or a model expression, such
    as sphere+sphere for a mixture of spheres of different radii, or
    sphere@hardsphere for concentrated solutions where the dilute approximation
    no longer applies.

    Use additional keywords for model parameters, tagged with *_pd*, *_pd_n*,
    *_pd_nsigma*, *_pd_type* to set polydispersity parameters, or *_M0*,
    *_mphi*, *_mtheta* for magnetic parameters.

    This is not intended for use when the same I(q) is evaluated many times
    with different parameter values. For that you should set up the model
    with `model = build_model(load_model_info(model_name))`, set up a data
    object to define q values and resolution, then use
    `calculator = DirectModel(data, model)` to set up a calculator, or
    `problem = bumps.FitProblem(sasmodels.bumps_model.Experiment(data, model))`
    to define a fit problem for uses with the bumps optimizer. Data can be
    loaded using the `sasdata` package, or use one of the empty data generators
    from `sasmodels.data`.

    Models are cached. Custom models will not be reloaded even if the
    underlying files have changed. If you are using this in a long running
    application then you will need to call
    `sasmodels.direct_model._model_cache.clear()` to reset the cache and force
    custom model reload.
    """
    from .data import Data1D, _as_numpy
    data = Data1D(x=q, dx=dq)
    def broadcast(v):
        return (
            None if v is None
            else np.full(len(q), v) if np.isscalar(v)
            else _as_numpy(v))
    data.dxl, data.dxw = broadcast(ql), broadcast(qw)
    return _direct_calculate(model, data, pars)

def Iqxy(model, qx, qy, dqx=None, dqy=None, **pars):
    """
    Compute I(qx, qy) for *model*. Resolution is *dqx* and *dqy*.
    See :func:`Iq` for details on model and parameters.
    """
    from .data import Data2D
    data = Data2D(x=qx, y=qy, dx=dqx, dy=dqy)
    return _direct_calculate(model, data, pars)

def Gxi(model, xi, **pars):
    """
    Compute SESANS correlation G' = G(xi) - G(0) for *model*.
    See :func:`Iq` for details on model and parameters.
    """
    from .data import empty_sesans
    data = empty_sesans(z=xi)
    return _direct_calculate(model, data, pars)

def main():
    # type: () -> None
    """
    Program to evaluate a particular model at a set of q values.
    """
    import sys

    if len(sys.argv) < 3:
        print("usage: python -m sasmodels.direct_model modelname (q|qx,qy) par=val ...")
        sys.exit(1)
    model = sys.argv[1]
    call = sys.argv[2].upper()
    pars = dict((k, (float(v) if not k.endswith("_pd_type") else v))
                for pair in sys.argv[3:]
                for k, v in [pair.split('=')])
    try:
        values = [float(v) for v in call.split(',')]
    except ValueError:
        values = []
    if len(values) == 1:
        q, = values
        dq = dqw = dql = None
        #dq = [q*0.05] # 5% pinhole resolution
        #dqw, dql = [q*0.05], [1.0] # 5% horizontal slit resolution
        print(Iq(model, [q], dq=dq, qw=dqw, ql=dql, **pars)[0])
        #print(Gxi(model, [q], **pars)[0])
    elif len(values) == 2:
        qx, qy = values
        dq = None
        #dq = [0.005] # 5% pinhole resolution at q = 0.1
        print(Iqxy(model, [qx], [qy], dqx=dq, dqy=dq, **pars)[0])
    else:
        print("use q or qx,qy")
        sys.exit(1)

def test_simple_interface():
    def near(value, target):
        """Close enough in single precision"""
        #print(f"value: {value}, target: {target}")
        return np.allclose(value, target, rtol=1e-6, atol=0, equal_nan=True)
    # Note: target values taken from running main() on parameters.
    # Resolution was 5% dq/q.
    pars = dict(radius=200, background=0)  # default background=1e-3, scale=1
    # simple sphere in 1D (perfect, pinhole, slit)
    perfect_target = 0.6190146273894904
    assert near(Iq('sphere', [0.1], **pars), [perfect_target])
    assert near(Iq('sphere', [0.1], dq=[0.005], **pars), [2.3009224683980215])
    assert near(Iq('sphere', [0.1], qw=[0.005], ql=[1.0], **pars), [0.3663431784535172])
    # simple sphere in 2D (perfect, pinhole)
    assert near(Iqxy('sphere', [0.1], [0.1], **pars), [1.1771532874802199])
    assert near(Iqxy('sphere', [0.1], [0.1], dqx=[0.005], dqy=[0.005], **pars), 
        [0.8167780778578667])
    # sesans (no background or scale)
    assert near(Gxi('sphere', [100], **pars), [-0.19146959126623486])
    # Check that single point sesans matches value in an array
    xi = np.logspace(1, 3, 100)
    y = Gxi('sphere', xi, **pars)
    for k in (0, len(xi)//5, len(xi)//2, len(xi)-1):
        ysingle = Gxi('sphere', [xi[k]], **pars)[0]
        print(f"SESANS point check {k}: xi={xi[k]:.1f} single={ysingle:.4f} vector={y[k]:.4f}")
        assert abs((ysingle-y[k])/y[k]) < 0.1, "SESANS point value not matching vector value within 10%"
    # magnetic 2D
    pars = dict(radius=200, sld_M0=3, sld_mtheta=30)
    assert near(Iqxy('sphere', [0.1], [0.1], **pars), [1.5577852226925908])
    # polydisperse 1D
    pars = dict(
        radius=200, radius_pd=0.1, radius_pd_n=15, radius_pd_nsigma=2.5,
        radius_pd_type="uniform")
    assert near(Iq('sphere', [0.1], **pars), [2.703169824954617])
    # background and scale
    background, scale = 1e-4, 0.1
    pars = dict(radius=200, background=background, scale=scale)
    assert near(Iq('sphere', [0.1], **pars), [perfect_target*scale + background])


if __name__ == "__main__":
    import logging
    logging.disable(logging.ERROR)
    main()
    #test_simple_interface()



#!/usr/bin/env python
"""
Generate the Gauss-Legendre integration points and save them as a C file.
"""
from __future__ import division, print_function

import numpy as np
from numpy.polynomial.legendre import leggauss

def gengauss(n, path):
    """
    Save the Gauss-Legendre integration points for length *n* into file *path*.
    """
    z, w = leggauss(n)

    # Make sure array size is a multiple of 4
    if n%4:
        array_size = n + (4 - n%4)
        z, w = [np.hstack((v, [0.]*(4-n%4))) for v in (z, w)]
    else:
        array_size = n

    with open(path, "w") as fid:
        fid.write("""\
// Generated by sasmodels.gengauss.gengauss(%d)

#ifdef GAUSS_N
# undef GAUSS_N
# undef GAUSS_Z
# undef GAUSS_W
#endif
#define GAUSS_N %d
#define GAUSS_Z Gauss%dZ
#define GAUSS_W Gauss%dWt

"""%(n, n, n, n))

        if array_size != n:
            fid.write("// Note: using array size %d so that it is a multiple of 4\n\n"%array_size)

        fid.write("constant double Gauss%dWt[%d]={\n"%(n, array_size))
        fid.write(",\n".join("\t% .15e"%v for v in w))
        fid.write("\n};\n")

        fid.write("constant double Gauss%dZ[%d]={\n"%(n, array_size))
        fid.write(",\n".join("\t% .15e"%v for v in z))
        fid.write("\n};")



#!/usr/bin/env python
"""
Small application to change theta, phi and psi from SasView 3.x models to the
new angle definition in SasView 4.x and above.

Usage: python explore/transform_angles.py theta phi psi
"""
from __future__ import print_function, division

import sys

import numpy as np
from numpy import pi, cos, sin, sqrt, exp, degrees, radians
from scipy.optimize import fmin

# Definition of rotation matrices comes from wikipedia:
#    https://en.wikipedia.org/wiki/Rotation_matrix#Basic_rotations
def Rx(angle):
    """Construct a matrix to rotate points about *x* by *angle* degrees."""
    a = radians(angle)
    R = [[1, 0, 0],
         [0, +cos(a), -sin(a)],
         [0, +sin(a), +cos(a)]]
    return np.array(R)

def Ry(angle):
    """Construct a matrix to rotate points about *y* by *angle* degrees."""
    a = radians(angle)
    R = [[+cos(a), 0, +sin(a)],
         [0, 1, 0],
         [-sin(a), 0, +cos(a)]]
    return np.array(R)

def Rz(angle):
    """Construct a matrix to rotate points about *z* by *angle* degrees."""
    a = radians(angle)
    R = [[+cos(a), -sin(a), 0],
         [+sin(a), +cos(a), 0],
         [0, 0, 1]]
    return np.array(R)


def transform_angles(theta, phi, psi, qx=0.1, qy=0.1):
    Rold = Rz(-psi)@Rx(theta)@Ry(-(90 - phi))
    cost = lambda p: np.linalg.norm(Rz(-p[2])@Ry(-p[0])@Rz(-p[1]) - Rold)
    result = fmin(cost, (theta, phi, psi))
    theta_p, phi_p, psi_p = result
    Rnew = Rz(-psi_p)@Ry(-theta_p)@Rz(-phi_p)

    print("old: theta, phi, psi =", ", ".join(str(v) for v in (theta, phi, psi)))
    print("new: theta, phi, psi =", ", ".join(str(v) for v in result))
    try:
        point = np.array([qx, qy, [0]*len(qx)])
    except TypeError:
        point = np.array([[qx],[qy],[0]])
    for p in point.T:
        print("q abc old for", p, (Rold@p.T).T)
        print("q abc new for", p, (Rnew@p.T).T)

if __name__ == "__main__":
    theta, phi, psi = (float(v) for v in sys.argv[1:])
    #transform_angles(theta, phi, psi)
    transform_angles(theta, phi, psi, qx=-0.017, qy=0.035)
